"""Color-related functions and data structures."""

from .color_types import AColor, AHexcode, Color, ColorTheme, Hexcode
from .colors import (
    ABLACK,
    ABLUE,
    ACYAN,
    AGREEN,
    AMAGENTA,
    ARED,
    AWHITE,
    AYELLOW,
    BLACK,
    BLUE,
    CYAN,
    GREEN,
    MAGENTA,
    NEPTUNE_PRIMARY_BG,
    NEPTUNE_PRIMARY_FG,
    NEPTUNE_THEME,
    RED,
    TRANSPARENT,
    WHITE,
    YELLOW,
    Neptune,
)
from .gradients import (
    darken_only,
    gradient,
    lerp_colors,
    lighten_only,
    rainbow_gradient,
)

__all__ = [
    "ABLACK",
    "ABLUE",
    "ACYAN",
    "AGREEN",
    "AMAGENTA",
    "ARED",
    "AWHITE",
    "AYELLOW",
    "BLACK",
    "BLACK_ON_BLACK",
    "BLUE",
    "CYAN",
    "NEPTUNE_THEME",
    "NEPTUNE_PRIMARY_BG",
    "NEPTUNE_PRIMARY_FG",
    "GREEN",
    "MAGENTA",
    "RED",
    "TRANSPARENT",
    "WHITE",
    "WHITE_ON_BLACK",
    "YELLOW",
    "AColor",
    "AHexcode",
    "Color",
    "ColorTheme",
    "Hexcode",
    "Neptune",
    "darken_only",
    "gradient",
    "lerp_colors",
    "lighten_only",
    "rainbow_gradient",
]



"""
A base for creating tree-like views. Tree views are composed of nodes that
can be selected and toggled open or closed.
"""

from collections.abc import Iterator
from typing import Self

from numpy.typing import NDArray

from .behaviors.button_behavior import ButtonBehavior, ButtonState
from .behaviors.themable import Themable
from .gadget import Cell, Gadget, Point, PosHint, Size, SizeHint
from .text import Text

__all__ = ["TreeView", "TreeViewNode", "ButtonState", "Point", "Size"]


class TreeViewNode(Themable, ButtonBehavior, Text):
    r"""
    A node of a :class:`TreeView`.

    Parameters
    ----------
    is_leaf : bool, default: True
        Whether node is a leaf node.
    always_release : bool, default: False
        Whether a mouse up event outside the button will trigger it.
    default_cell : NDArray[Cell] | str, default: " "
        Default cell of text canvas.
    alpha : float, default: 0.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    root_node : TreeViewNode
        Root node of tree.
    always_release : bool
        Whether a mouse up event outside the button will trigger it.
    state : ButtonState
        Current button state. One of `NORMAL`, `HOVER`, `DOWN`.
    canvas : NDArray[Cell]
        The array of characters for the gadget.
    default_cell : NDArray[Cell]
        Default cell of text canvas.
    default_fg_color : Color
        Foreground color of default cell.
    default_bg_color : Color
        Background color of default cell.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    iter_open_nodes()
        Yield all open descendent nodes.
    add_node(node: TreeViewNode)
        Add a child node.
    remove_node(node: TreeViewNode)
        Remove a child node.
    toggle()
        Toggle node open or closed.
    select()
        Select this node.
    unselect()
        Unselect this node.
    get_color()
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    update_normal()
        Paint the normal state.
    update_hover()
        Paint the hover state.
    update_down()
        Paint the down state.
    on_release()
        Triggered when a button is released.
    add_border(style="light", ...)
        Add a border to the gadget.
    add_syntax_highlighting(lexer, style)
        Add syntax highlighting to current text in canvas.
    add_str(str, pos, ...)
        Add a single line of text to the canvas.
    set_text(text, ...)
        Resize gadget to fit text, erase canvas, then fill canvas with text.
    clear()
        Fill canvas with default cell.
    shift(n=1)
        Shift content in canvas up (or down in case of negative `n`).
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        is_leaf=True,
        always_release: bool = False,
        default_cell: NDArray[Cell] | str = " ",
        alpha: float = 0.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.is_leaf: bool = is_leaf
        self.is_open: bool = False
        self.is_selected: bool = False
        self.parent_node: Self | None = None
        self.child_nodes: list[Self] = []
        self.level: int = -1
        super().__init__(
            always_release=always_release,
            default_cell=default_cell,
            alpha=alpha,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

    def _repaint(self):
        if self.is_selected:
            fg = self.get_color("menu_item_selected_fg")
            bg = self.get_color("menu_item_selected_bg")
        elif self.button_state == "normal":
            fg = self.get_color("primary_fg")
            bg = self.get_color("primary_bg")
        else:
            fg = self.get_color("menu_item_hover_fg")
            bg = self.get_color("menu_item_hover_bg")
        self.canvas["fg_color"] = fg
        self.canvas["bg_color"] = bg

    def on_size(self):
        """Repaint tree on resize."""
        super().on_size()
        self._repaint()

    def update_theme(self):
        """Paint the gadget with current theme."""
        self._repaint()

    def update_normal(self):
        """Paint the normal state."""
        self._repaint()

    def update_hover(self):
        """Paint the hover state."""
        self._repaint()

    def update_down(self):
        """Paint the down state."""
        self._repaint()

    @property
    def root_node(self) -> Self:
        """Root node of tree."""
        if self.parent_node is None:
            return self
        return self.parent_node.root_node

    def iter_open_nodes(self) -> Iterator[Self]:
        """
        Yield all open descendent nodes.

        Yields
        ------
        TreeViewNode
            A descendent open node.
        """
        for child in self.child_nodes:
            yield child

            if child.is_open:
                yield from child.iter_open_nodes()

    def add_node(self, node: Self):
        """
        Add a child node.

        Parameters
        ----------
        node : TreeViewNode
            The node to add.
        """
        self.child_nodes.append(node)

        node.level = self.level + 1
        node.parent_node = self

    def remove_node(self, node: Self):
        """
        Remove a child node.

        Parameters
        ----------
        node : TreeViewNode
            The node to remove.
        """
        self.child_nodes.remove(node)

        node.level = -1
        node.parent_node = None

    def _toggle_update(self):
        """Update state after :meth:`toggle` is called."""

    def toggle(self):
        """Toggle node open or closed."""
        if not self.is_leaf:
            self.is_open = not self.is_open
            self._toggle_update()
            self.root_node.tree_view.update_tree_layout()

    def select(self):
        """Select node."""
        if self.root_node.tree_view.selected_node is not None:
            self.root_node.tree_view.selected_node.unselect()

        self.is_selected = True
        self.root_node.tree_view.selected_node = self
        self._repaint()

    def unselect(self):
        """Unselect node."""
        self.is_selected = False
        self.root_node.tree_view.selected_node = None
        self._repaint()

    def on_release(self):
        """Select and toggle node on release."""
        self.select()
        self.toggle()


class TreeView(Gadget):
    r"""
    Base for creating tree-like views.

    Parameters
    ----------
    root_node : TreeViewNode
        Root node of tree view.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    root_node : TreeViewNode
        Root node of tree view
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    update_tree_layout()
        Update tree layout after a child node is toggled open or closed.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        root_node: TreeViewNode,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.selected_node = None
        self.root_node = root_node
        root_node.tree_view = self

        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        root_node.toggle()

    def update_tree_layout(self) -> None:
        """Update tree layout after a child node is opened or closed."""



"""An animation gadget."""

import asyncio
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Self

import numpy as np
from numpy.typing import NDArray

from ..colors import TRANSPARENT, AColor
from ..texture_tools import read_texture, resize_texture
from .graphics import (
    Blitter,
    Graphics,
    Interpolation,
    Point,
    PosHint,
    Size,
    SizeHint,
    scale_geometry,
)

__all__ = ["Animation", "Interpolation", "Point", "Size"]


class Animation(Graphics):
    r"""
    An animation gadget.

    Parameters
    ----------
    path : Path | None, default: None
        Path to directory of images for frames in the animation (loaded in lexographical
        order of filenames).
    frame_durations : float | Sequence[float], default: 1/12
        Time each frame is displayed. If a sequence is provided, it's length should be
        equal to number of frames.
    loop : bool, default: True
        Whether to restart animation after last frame.
    reverse : bool, default: False
        Whether to play animation in reverse.
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    blitter : Blitter, default: "half"
        Determines how graphics are rendered.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    frames : list[Image]
        Frames of the animation.
    frame_durations : list[int | float]
        Time each frame is displayed.
    loop : bool
        Whether to animation is restarted after last frame.
    reverse : bool
        Whether to animation is played in reverse.
    texture : NDArray[np.uint8]
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    blitter : Blitter
        Determines how graphics are rendered.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    play()
        Play the animation. Returns a task.
    pause()
        Pause the animation
    stop()
        Stop the animation and reset current frame.
    from_textures(textures, ...)
        Create an :class:`Animation` from an iterable of uint8 RGBA numpy array.
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        path: Path | None = None,
        frame_durations: float | Sequence[float] = 1 / 12,
        loop: bool = True,
        reverse: bool = False,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        blitter: Blitter = "half",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.frames: list[NDArray[np.uint8]]
        """Frames of the animation."""
        # Set in `on_size`
        self._sized_frames: list[NDArray[np.uint8]]
        """Resized frames of the animation."""

        if path is None:
            self.frames = []
        else:
            paths = sorted(path.iterdir(), key=lambda file: file.name)
            self.frames = [read_texture(path) for path in paths]

        self.frame_durations: Sequence[float]
        """Time each frame is displayed."""

        nframes = len(self.frames)
        if isinstance(frame_durations, Sequence):
            if len(frame_durations) != nframes:
                raise ValueError(
                    "number of frames doesn't match number of frame durations"
                )
            self.frame_durations = frame_durations
        else:
            self.frame_durations = [frame_durations] * nframes

        super().__init__(
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            blitter=blitter,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.loop = loop
        self.reverse = reverse
        self._i = len(self.frames) - 1 if self.reverse else 0
        self._animation_task = None

    @property
    def texture(self) -> NDArray[np.uint8]:
        """uint8 RGBA color array."""
        if self._i < len(self.frames):
            return self._sized_frames[self._i]
        return self._texture

    @texture.setter
    def texture(self, texture: NDArray[np.uint8]) -> None:
        self._texture = texture

    def on_remove(self) -> None:
        """Pause animation."""
        self.pause()
        super().on_remove()

    def on_size(self) -> None:
        """Update size of all frames on resize."""
        size = scale_geometry(self._blitter, self._size)
        self._texture = np.full((*size, 4), self.default_color, np.uint8)
        self._sized_frames = [
            resize_texture(texture, size, self.interpolation) for texture in self.frames
        ]

    async def _play_animation(self):
        while self.frames:
            try:
                await asyncio.sleep(self.frame_durations[self._i])
            except asyncio.CancelledError:
                return

            if self.reverse:
                self._i -= 1
                if self._i < 0:
                    self._i = len(self.frames) - 1
                    if not self.loop:
                        return
            else:
                self._i += 1
                if self._i >= len(self.frames):
                    self._i = 0

                    if not self.loop:
                        return

    def play(self) -> asyncio.Task:
        """
        Play animation.

        Returns
        -------
        asyncio.Task
            The task that plays the animation.
        """
        self.pause()

        if self._i == 0 and self.reverse:
            self._i = len(self.frames) - 1
        elif self._i == len(self.frames) - 1 and not self.reverse:
            self._i = 0

        self._animation_task = asyncio.create_task(self._play_animation())
        return self._animation_task

    def pause(self) -> None:
        """Pause animation."""
        if self._animation_task is not None:
            self._animation_task.cancel()

    def stop(self) -> None:
        """Stop the animation and reset current frame."""
        self.pause()
        self._i = len(self.frames) - 1 if self.reverse else 0

    @classmethod
    def from_textures(
        cls,
        textures: Iterable[NDArray[np.uint8]],
        *,
        frame_durations: float | Sequence[float] = 1 / 12,
        loop: bool = True,
        reverse: bool = False,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        blitter: Blitter = "half",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ) -> Self:
        """
        Create an :class:`Animation` from an iterable of uint8 RGBA numpy array.

        Parameters
        ----------
        textures : Iterable[NDArray[np.uint8]]
            An iterable of RGBA textures that will be the frames of the animation.
        frame_durations : float | Sequence[float], default: 1/12
            Time each frame is displayed. If a sequence is provided, it's length should
            be equal to number of frames.
        loop : bool, default: True
            Whether to restart animation after last frame.
        reverse : bool, default: False
            Whether to play animation in reverse.
        default_color : AColor, default: AColor(0, 0, 0, 0)
            Default texture color.
        alpha : float, default: 1.0
            Transparency of gadget.
        interpolation : Interpolation, default: "linear"
            Interpolation used when gadget is resized.
        blitter : Blitter, default: "half"
            Determines how graphics are rendered.
        size : Size, default: Size(10, 10)
            Size of gadget.
        pos : Point, default: Point(0, 0)
            Position of upper-left corner in parent.
        size_hint : SizeHint | None, default: None
            Size as a proportion of parent's height and width.
        pos_hint : PosHint | None, default: None
            Position as a proportion of parent's height and width.
        is_transparent : bool, default: True
            Whether gadget is transparent.
        is_visible : bool, default: True
            Whether gadget is visible. Gadget will still receive input events if not
            visible.
        is_enabled : bool, default: True
            Whether gadget is enabled. A disabled gadget is not painted and doesn't
            receive input events.

        Returns
        -------
        Animation
            A new animation gadget.
        """
        frames = list(textures)
        nframes = len(frames)
        if isinstance(frame_durations, Sequence):
            if len(frame_durations) != nframes:
                raise ValueError(
                    "number of frames doesn't match number of frame durations"
                )
        else:
            frame_durations = [frame_durations] * nframes

        animation = cls(
            loop=loop,
            reverse=reverse,
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            blitter=blitter,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        animation.frames = frames
        animation.frame_durations = frame_durations
        animation.on_size()
        return animation



"""A sparkline gadget."""

from collections.abc import Sequence
from numbers import Real

import numpy as np
from numpy.typing import NDArray

from ..colors import NEPTUNE_PRIMARY_BG, NEPTUNE_PRIMARY_FG, Color, lerp_colors
from ..terminal.events import MouseEvent
from ..text_tools import smooth_vertical_bar
from .cursor import Cursor
from .gadget import Gadget, Point, PosHint, Size, SizeHint
from .text import Text, add_text

__all__ = ["Sparkline", "Point", "Size"]

DEFAULT_MIN_COLOR = Color.from_hex("1b244b")
DEFAULT_MAX_COLOR = Color.from_hex("4d67ff")
DEFAULT_HIGHLIGHT_COLOR = Color.from_hex("7281ff")


def _get_float_text(value: float) -> str:
    text = f"{value:8.2f}"
    if len(text) > 8:
        return f"{value:3.2e}"
    return text


class _Tooltip(Text):
    def on_mouse(self, mouse_event: MouseEvent) -> bool | None:
        self.is_enabled = self.selector.is_enabled = False


class Sparkline(Gadget):
    r"""
    A sparkline gadget for displaying sequential data.

    Parameters
    ----------
    data : Sequence[Real] | None, default: None
        Data for the sparkline.
    min_color : Color, default: DEFAULT_MIN_COLOR
        Color of minimum value of the sparkline.
    max_color : Color, default: DEFAULT_MAX_COLOR
        Color of the maximum value of the sparkline.
    highlight_color : Color, default: DEFAULT_HIGHLIGHT_COLOR
        Color of highlighted value of the sparkline.
    bg_color : Color, default: NEPTUNE_PRIMARY_BG
        Background color of gadget.
    show_tooltip : bool, default: True
        Whether to show tooltip.
    tooltip_fg_color : Color, default: NEPTUNE_PRIMARY_FG
        Foreground color of tooltip.
    tooltip_bg_color : Color, default: NEPTUNE_PRIMARY_BG
        Background color of tooltip.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    data : NDArray[np.float64]
        Data for the sparkline. This can be set with a `Sequence[Real]` or `None`, but
        will be converted into a (possibly empty) numpy array.
    min_color : Color
        Color of minimum value of the sparkline.
    max_color : Color
        Color of the maximum value of the sparkline.
    highlight_color : Color
        Color of highlighted value of the sparkline.
    bg_color : Color
        Background color of gadget.
    show_tooltip : bool
        Whether to show tooltip.
    tooltip_fg_color : Color
        Foreground color of tooltip.
    tooltip_bg_color : Color
        Background color of tooltip.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        data: Sequence[Real] | None = None,
        min_color: Color = DEFAULT_MIN_COLOR,
        max_color: Color = DEFAULT_MAX_COLOR,
        highlight_color: Color = DEFAULT_HIGHLIGHT_COLOR,
        bg_color: Color = NEPTUNE_PRIMARY_BG,
        show_tooltip: bool = True,
        tooltip_fg_color: Color = NEPTUNE_PRIMARY_FG,
        tooltip_bg_color: Color = NEPTUNE_PRIMARY_BG,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._sparkline = Text(size_hint={"height_hint": 1.0, "width_hint": 1.0})
        self._selector = Cursor(reverse=False, size_hint={"height_hint": 1.0})
        self._tooltip = _Tooltip(size=(7, 18), is_enabled=False)

        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self._tooltip.selector = self._selector
        self._tooltip.add_border(style="thick")
        add_text(
            self._tooltip.canvas[1:, 2:],
            "*Start:*\n*Stop:*\n*Min:*\n*Max:*\n*Mean:*",
            markdown=True,
        )
        self.add_gadget(self._sparkline)
        self._sparkline.add_gadget(self._selector)

        self.data = data
        self._min_color = min_color
        """Color of minimum value of the sparkline."""
        self._max_color = max_color
        """Color of the maximum value of the sparkline."""
        self.highlight_color = highlight_color
        self.bg_color = bg_color
        self.show_tooltip = show_tooltip
        self.tooltip_fg_color = tooltip_fg_color
        self.tooltip_bg_color = tooltip_bg_color

        # Following are set in `_build_sparkline`:
        self._walls: NDArray[np.float64]
        """
        Boundaries for each bin. The `i`th bin starts at ``self._walls[i]`` and stops at
        ``self._walls[i + 1]``.
        """
        self._mins: NDArray[np.float64]
        """Mininum of each bin."""
        self._maxs: NDArray[np.float64]
        """Maximum of each bin."""
        self._means: NDArray[np.float64]
        """Arithmetic mean of each bin."""

    @property
    def bg_color(self) -> Color:
        """Background color of gadget."""
        return self._selector.bg_color

    @bg_color.setter
    def bg_color(self, bg_color: Color):
        self._selector.bg_color = bg_color
        self._sparkline.default_bg_color = bg_color
        self._sparkline.canvas["bg_color"] = bg_color

    @property
    def highlight_color(self) -> Color:
        """Color of highlighted value of the sparkline."""
        return self._selector.fg_color

    @highlight_color.setter
    def highlight_color(self, highlight_color: Color):
        self._selector.fg_color = highlight_color

    @property
    def show_tooltip(self) -> bool:
        """Whether to show tooltip."""
        return self._show_tooltip

    @show_tooltip.setter
    def show_tooltip(self, show_tooltip: bool):
        self._show_tooltip = show_tooltip
        self._tooltip.is_enabled = False
        self._selector.is_enabled = False

    @property
    def tooltip_fg_color(self) -> Color:
        """Foreground color of tooltip."""
        return self._tooltip.default_fg_color

    @tooltip_fg_color.setter
    def tooltip_fg_color(self, tooltip_fg_color: Color):
        self._tooltip.default_fg_color = tooltip_fg_color
        self._tooltip.canvas["fg_color"] = tooltip_fg_color

    @property
    def tooltip_bg_color(self) -> Color:
        """Background color of tooltip."""
        return self._tooltip.default_bg_color

    @tooltip_bg_color.setter
    def tooltip_bg_color(self, tooltip_bg_color: Color):
        self._tooltip.default_bg_color = tooltip_bg_color
        self._tooltip.canvas["bg_color"] = tooltip_bg_color

    @property
    def min_color(self) -> Color:
        """Color of minimum value of the sparkline."""
        return self._min_color

    @min_color.setter
    def min_color(self, min_color: Color):
        self._min_color = min_color
        self._build_sparkline()

    @property
    def max_color(self) -> Color:
        """Color of the maximum value of the sparkline."""
        return self._max_color

    @max_color.setter
    def max_color(self, max_color: Color):
        self._max_color = max_color
        self._build_sparkline()

    @property
    def data(self) -> NDArray[np.float64]:
        """Data for the sparkline."""
        return self._data

    @data.setter
    def data(self, data: Sequence[Real] | None):
        self._data = np.array([]) if data is None else np.array(data, float)
        self._build_sparkline()

    def on_add(self):
        """Add tooltip to root and build sparkline on add."""
        super().on_add()
        self.root.add_gadget(self._tooltip)
        self._build_sparkline()

    def on_remove(self):
        """Remove tooltip from root on remove."""
        self.root.remove_gadget(self._tooltip)

    def on_size(self):
        """Rebuild sparkline on resize."""
        super().on_size()
        self._build_sparkline()

    def _build_sparkline(self):
        if not self.root:
            return

        self._selector.is_enabled = False
        self._tooltip.is_enabled = False

        if len(self.data) <= self.width:
            self._walls = np.arange(len(self.data) + 1)
            self._means = self._mins = self._maxs = self.data
            bin_proportions = (self.data - self.data.min(initial=0)) / (
                self.data.max(initial=0) - self.data.min(initial=0)
            )
        else:
            nbins = self.width
            self._walls = np.zeros(nbins + 1, int)
            self._means = np.zeros(nbins)
            self._mins = np.zeros(nbins)
            self._maxs = np.zeros(nbins)
            bin_width = len(self.data) / nbins
            x = 0
            for i in range(nbins):
                start = self._walls[i]
                stop = round(x + bin_width)
                x += bin_width
                self._walls[i + 1] = stop
                self._mins[i] = self.data[start:stop].min()
                self._maxs[i] = self.data[start:stop].max()
                self._means[i] = np.mean(self.data[start:stop])
            bin_proportions = (self._means - self._means.min()) / (
                self._means.max() - self._means.min()
            )

        self._sparkline.clear()
        for i, bin_proportion in enumerate(bin_proportions):
            smooth_bar = smooth_vertical_bar(self.height, bin_proportion)
            self._sparkline.chars[::-1][: len(smooth_bar), i] = smooth_bar
            self._sparkline.canvas["fg_color"][:, i] = lerp_colors(
                self.min_color, self.max_color, bin_proportion
            )

    def on_mouse(self, mouse_event: MouseEvent) -> bool | None:
        """Show tooltip and highlight column on mouse collision."""
        if not self.collides_point(mouse_event.pos):
            return

        _, x = self.to_local(mouse_event.pos)
        if self.show_tooltip and len(self._means) > x:
            start = f"{self._walls[x]:8d}"
            stop = f"{self._walls[x + 1]:8d}"
            min_ = _get_float_text(self._mins[x])
            max_ = _get_float_text(self._maxs[x])
            mean = _get_float_text(self._means[x])

            add_text(
                self._tooltip.canvas[1:, 8:],
                f"{start}\n{stop}\n{min_}\n{max_}\n{mean}",
            )

            tth, ttw = self._tooltip.size
            my, mx = mouse_event.pos
            rh, rw = self.root.size

            # Keep the tooltip inbounds:
            if my + tth + 1 < rh:
                tty = my + 1
            elif my - tth - 1 >= 0:
                tty = my - tth - 1
            else:
                tty = 0

            if mx + ttw + 1 < rw:
                ttx = mx + 1
            elif mx - ttw - 1 >= 0:
                ttx = mx - ttw - 1
            else:
                ttx = 0

            self._tooltip.pos = tty, ttx
            self._tooltip.pull_to_front()
            self._tooltip.is_enabled = True

            self._selector.x = x
            self._selector.is_enabled = True

            return True



"""A shadow caster gadget."""

from collections.abc import Callable
from itertools import product
from typing import Literal

import numpy as np
from numpy.typing import NDArray

from ..colors import AWHITE, BLACK, TRANSPARENT, AColor, Color
from ._shadow_casting import cast_shadows
from .graphics import (
    Blitter,
    Graphics,
    Interpolation,
    Point,
    PosHint,
    Size,
    SizeHint,
    clamp,
)

__all__ = ["ShadowCaster", "Restrictiveness", "Point", "Size"]

AGRAY = AColor(50, 50, 50)
QUADS = tuple(product((1, -1), (1, -1), (False, True)))

Restrictiveness = Literal["permissive", "moderate", "restrictive"]
"""
The restrictiveness of the shadow caster.

``Restrictiveness`` is one of "permissive", "moderate", "restrictive".

For "permissive", any interval is visible as long as any of it's start, center, or end
points are visible. For "moderate", the center and either end must be visible. For
"restrictive", all points in the interval must be visible.
"""


def default_light_decay(x: float) -> float:
    if x == 0:
        return 1
    return 1 / x


class ShadowCaster(Graphics):
    r"""
    A restrictive precise angle shadowcaster.

    :meth:`cast_shadows` must be called to generate or update :attr:`texture`.

    Light decay distance and :attr:`radius` are calculated from the point of view of the
    ``caster_map`` and does not depend on the size of the camera or the size of the
    caster. As a consequence, changing the blitter should only change the resolution of
    the caster.

    ``light_coords`` and ``light_colors`` must be same length.

    Parameters
    ----------
    caster_map : NDArray[np.uint8]
        A 2-d map. Non-zero values are walls.
    camera_pos : Point
        Position of camera in map.
    camera_size : Size
        Size of camera. Determines how much of the map is visible.
    tile_colors : list[AColor] | None, default: None
        A value ``n`` in the map will be colored ``tile_colors[n]``. If ``None``,
        ``tile_colors`` will be set to ``[AGRAY, AWHITE]``.
    light_coords : list[tuple[float, float]] | None, default: None
        A list of coordinates for all light sources on the map.
    light_colors : list[Color] | None, default: None
        A list of colors for all light sources on the map.
    ambient_light : Color, default: BLACK
        Color of ambient light. Default is no light.
    restrictiveness : Restrictiveness, default: "permissive"
        Restrictiveness of casting algorithm.
    radius : int, default: 20
        Max visible radius from a light source.
    smoothing : float, default: 1.0 / 3.0
        Smoothness of shadow edges. This value will be clamped between ``0`` and ``1``.
    not_visible_blocks : bool, default: True
        Whether all not-visible cells will be treated as opaque.
    light_decay : Callable[[float], float], default: default_light_decay
        The strength of light as a function of distance from light origin. The default
        decay is ``1 / distance``.
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    blitter : Blitter, default: "half"
        Determines how graphics are rendered.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    caster_map : NDArray[np.uint8]
        A 2-d map. Non-zero values are walls.
    camera_pos : Point
        Position of camera in map.
    camera_size : Size
        Size of camera. Determines how much of the map is visible.
    tile_colors : list[AColor]
        A value ``n`` in the map will be colored ``tile_colors[n]``.
    light_coords : list[tuple[float, float]] | None
        A list of coordinates for all light sources on the map.
    light_colors : list[Color] | None
        A list of colors for all light sources on the map.
    ambient_light : Color
        Color of ambient light.
    restrictiveness : Restrictiveness
        Restrictiveness of casting algorithm.
    radius : int
        Max visible radius from a light source.
    smoothing : float
        Smoothness of shadow edges.
    not_visible_blocks : bool
        Whether all not-visible cells will be treated as opaque.
    light_decay : Callable[[float], float]
        The strength of light as a function of distance from origin.
    texture : NDArray[np.uint8]
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    blitter : Blitter
        Determines how graphics are rendered.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    cast_shadows()
        Update texture by shadow casting all light sources.
    to_map_coords(point)
        Convert a point in the gadget's local coordinates to a point in the map's
        coordinates.
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        caster_map: NDArray[np.uint8],
        camera_pos: Point,
        camera_size: Size,
        tile_colors: list[AColor] | None = None,
        light_coords: list[tuple[float, float]] | None = None,
        light_colors: list[Color] | None = None,
        ambient_light: Color = BLACK,
        restrictiveness: Restrictiveness = "permissive",
        radius: int = 20,
        smoothing: float = 1.0 / 3.0,
        not_visible_blocks: bool = True,
        light_decay: Callable[[float], float] = default_light_decay,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        blitter: Blitter = "half",
        interpolation: Interpolation = "linear",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = True,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            default_color=default_color,
            alpha=alpha,
            interpolation=interpolation,
            blitter=blitter,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.caster_map = caster_map
        self.camera_pos = camera_pos
        self.camera_size = camera_size
        self.tile_colors = tile_colors or [AGRAY, AWHITE]
        self.light_coords = light_coords or []
        self.light_colors = light_colors or []
        self.ambient_light = ambient_light
        self.restrictiveness = restrictiveness
        self.radius = radius
        self.smoothing = clamp(smoothing, 0.0, 1.0)
        self.not_visible_blocks = not_visible_blocks
        self.light_decay = light_decay

    def on_size(self):
        """Resize ``texture`` and ``_light_intensity`` buffer on size."""
        super().on_size()
        h, w, _ = self.texture.shape
        self._light_intensity = np.empty((h, w, 3), float)

    def cast_shadows(self):
        """Update texture by shadow casting all light sources."""
        self._light_intensity[:] = self.ambient_light
        cast_shadows(
            self.texture,
            self._light_intensity,
            self.caster_map,
            self.camera_pos,
            self.camera_size,
            self.tile_colors,
            self.light_coords,
            self.light_colors,
            self.restrictiveness,
            self.radius,
            self.smoothing,
            self.not_visible_blocks,
            self.light_decay,
        )

    def to_map_coords(self, point: Point) -> tuple[float, float]:
        """
        Convert a point in the gadget's local coordinates to a point in the map's
        coordinates.

        Parameters
        ----------
        point : Point
            Point in local coordinates.

        Returns
        -------
        tuple[float, float]
            The coordinates of the point in the map.
        """
        y, x = point
        h, w = self.size

        cy, cx = self.camera_pos
        ch, cw = self.camera_size

        return ch / h * y + cy, cw / w * x + cx



"""An interactive python console gadget."""

from __future__ import annotations

import asyncio
import builtins
import sys
import time
from code import InteractiveInterpreter
from contextlib import redirect_stderr, redirect_stdout
from io import StringIO

from wcwidth import wcswidth

from ..terminal.events import KeyEvent, PasteEvent
from .behaviors.focusable import Focusable
from .behaviors.themable import Themable
from .gadget import Gadget, Point, PosHint, Size, SizeHint
from .pane import Pane
from .scroll_view import ScrollView
from .text import Text
from .textbox import Textbox

__all__ = ["Console", "Point", "Size"]

PROMPT_1 = ">>> "
PROMPT_2 = "... "
DEFAULT_BANNER = f"""\
Welcome to the batgrl interactive console!
The root gadget is "root" and the running app is "app".
Python {sys.version} on {sys.platform}
Type "help", "copyright", "credits" or "license" for more information."""


class _InteractiveConsole(InteractiveInterpreter):
    """An custom interactive interpreter."""

    def __init__(self, console_gadget):
        super().__init__()
        self.filename = "<batgrl console>"
        self.src_buffer = []
        self.console_gadget: Console = console_gadget
        self.input_buffer = None
        """Result of replaced stdin.read or stdin.readline."""
        self.output_buffer = StringIO()
        """Replaces stdout while executing code."""
        self.exec_in_thread: bool = False
        """Whether code is executed in a separate thread."""

        def read(size=-1):
            return self.input()

        self.stdin = type("stdin", (), {"read": read, "readline": read})()
        """Replaces sys.stdin while executing code."""

    def flush(self):
        """Write the output buffer to console gadget's output."""
        text = self.output_buffer.getvalue()
        self.output_buffer.truncate(0)
        self.console_gadget._add_text_to_output(text)

    def input(self, prompt=""):
        """Replace `builtins.input` while executing code."""
        if not self.exec_in_thread:
            print("set `exec_in_thread` to True to enable `input`")

        *lines, last_line = str(prompt).split("\n")
        self.output_buffer.write("\n".join(lines))
        self.flush()

        if not self.exec_in_thread:
            print(last_line)
            return ""

        self.console_gadget._input_mode = True
        self.console_gadget._prompt.set_text(last_line)

        try:
            # Blocking
            while self.input_buffer is None:
                time.sleep(0.01)
            return self.input_buffer
        finally:
            self.input_buffer = None
            self.console_gadget._input_mode = False
            self.console_gadget._prompt.set_text(PROMPT_1)

    def exec(self, code):
        old_input, builtins.input = builtins.input, self.input
        old_stdin, sys.stdin = sys.stdin, self.stdin

        with (
            redirect_stdout(self.output_buffer),
            redirect_stderr(self.output_buffer),
        ):
            try:
                exec(code, self.locals)
            except SystemExit:
                raise
            except:  # noqa
                self.showtraceback()
            finally:
                builtins.input = old_input
                sys.stdin = old_stdin
                self.flush()

    def runcode(self, code):
        if self.exec_in_thread:
            loop = asyncio.get_running_loop()
            loop.run_in_executor(None, self.exec, code)
        else:
            self.exec(code)

    def push(self, line):
        self.src_buffer.append(line)
        source = "\n".join(self.src_buffer)

        with redirect_stderr(self.output_buffer):
            more = self.runsource(source)

        self.flush()

        if not more:
            self.src_buffer.clear()
            self.console_gadget._prompt.set_text(PROMPT_1)
        else:
            self.console_gadget._prompt.set_text(PROMPT_2)


class _ConsoleTextbox(Textbox):
    """A custom textbox that grows and shrinks with its input."""

    @property
    def console(self) -> Console:
        return self.parent.parent.parent

    @property
    def cursor(self) -> int:
        return self._cursor.x

    @cursor.setter
    def cursor(self, cursor: int):
        self._cursor.x = cursor

        if self.parent is None:
            return

        y, x = self.pos
        self.console._scroll_view.scroll_to_rect((y, cursor + x))

        if self.is_selecting:
            self._selection_end = self.cursor

        self._highlight_selection()

    def on_size(self):
        if self.right >= self.console._min_line_length:
            self.console._container.width = self.right
        else:
            self.console._container.width = self.console._min_line_length
        self.cursor = self.cursor

    def _del_text(self, start: int, end: int):
        result = super()._del_text(start, end)
        self.width = self._box.width = self._line_width + 1
        return result

    def _add_text(self, x: int, text: str):
        result = super()._add_text(x, text)
        self.width = self._box.width = self._line_width + 1
        return result

    def _tab(self):
        self._move_undo_buffer_to_stack()
        undos = []
        if undo := self.delete_selection():
            undos.append(undo)
        undos.append(self._add_text(self.cursor, "    "))
        self._undo_stack.append(undos)
        self._redo_stack.clear()

    def update_theme(self):
        super().update_theme()
        self._cursor.fg_color = None
        self._cursor.bg_color = None
        self._cursor.reverse = True

    def on_key(self, key_event: KeyEvent) -> bool | None:
        if (
            key_event.key == "tab"
            and not key_event.alt
            and not key_event.ctrl
            and not key_event.shift
        ):
            self._tab()
            return True
        if (
            key_event.key == "left"
            and not key_event.alt
            and not key_event.ctrl
            and not key_event.shift
            and self.cursor == 0
        ):
            self.console._scroll_view.horizontal_proportion = 0
            return True
        return super().on_key(key_event)

    def on_paste(self, paste_event: PasteEvent) -> bool | None:
        *lines, last = paste_event.paste.split("\n")
        for line in lines:
            super().on_paste(PasteEvent(line))
            self.enter_callback(self)
        super().on_paste(PasteEvent(last))
        return True

    def on_focus(self):
        pass

    def on_blur(self):
        # A hack to prevent textbox from losing focus when console is clicked:
        if self.parent and self.console.is_focused:
            self.focus()


def _enter_callback(textbox: Textbox):
    """Add input to history then push input on enter for _ConsoleTextbox."""
    text = textbox.text.rstrip()
    textbox.text = ""
    textbox.width = 1
    console: Console = textbox.parent.parent.parent
    console._add_text_to_output(text, with_prompt=True)

    if (
        isinstance(console._history_index, float)
        or console._history_index >= -1
        or text != console._history[console._history_index]
    ):
        console._history_index = 0

    if text and (len(console._history) == 0 or console._history[-1] != text):
        console._history.append(text)
        console._history_index -= 0.5
    elif console._history_index < 0:
        console._history_index += 0.5

    if console._input_mode:
        console._console.input_buffer = text
    else:
        console._console.push(text)


class _Prompt(Text):
    def set_text(self, text: str, **kwargs):
        self._text = text
        super().set_text(text, **kwargs)


class _AlmostPane(Pane):
    def _render(self, canvas, graphics, kind):
        console: Console = self.parent.parent
        self._region -= console._input._region
        super()._render(canvas, graphics, kind)


class Console(Themable, Focusable, Gadget):
    r"""
    An interactive python console gadget.

    Parameters
    ----------
    banner : str | None, default: None
        The banner to print in the console before first interaction. If not provided, a
        default banner is printed.
    exec_in_thread : bool, default: False
        Whether code is executed in a separate thread. Code not executed in a separated
        thread can block the event loop.
    size : Size, default: Size(10, 10)
        Size of gadget.
    alpha : float, default: 1.0
        Transparency of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    exec_in_thread : bool
        Whether code is executed in a separate thread.
    alpha : float
        Transparency of gadget.
    is_focused : bool
        Whether gadget has focus.
    any_focused : bool
        Whether any gadget has focus.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    get_color()
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    focus()
        Focus gadget.
    blur()
        Un-focus gadget.
    focus_next()
        Focus next focusable gadget.
    focus_previous()
        Focus previous focusable gadget.
    on_focus()
        Update gadget when it gains focus.
    on_blur()
        Update gadget when it loses focus.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    # TODO: Add a max output lines option.
    # TODO: Add a max input history option.
    # ? Maybe add some shell commands.

    def __init__(
        self,
        banner: str | None = None,
        exec_in_thread: bool = False,
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._output = Text(
            pos=(-1, 0), size=(1, 1), alpha=0, is_visible=False, is_transparent=True
        )
        self._prompt = _Prompt(alpha=0, is_transparent=True)
        self._input = _ConsoleTextbox(
            size=(1, 1), enter_callback=_enter_callback, is_transparent=is_transparent
        )
        self._container = Gadget(size=(1, 1), is_transparent=True)

        self._scroll_view = ScrollView(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
            dynamic_bars=True,
            arrow_keys_enabled=False,
            is_transparent=is_transparent,
        )
        # Replace scroll view background with a pane that doesn't paint under _input.
        self._scroll_view.remove_gadget(self._scroll_view._background)
        self._scroll_view._background = _AlmostPane(
            size_hint={"height_hint": 1.0, "width_hint": 1.0}
        )
        self._scroll_view.add_gadget(self._scroll_view._background)
        self._scroll_view.children.insert(0, self._scroll_view.children.pop())
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self._console = _InteractiveConsole(self)
        self._input_mode: bool = False
        self._history = []
        self._history_index: float | int = 0
        """
        Current index into history.

        If this value is a float then it is halfway between two integer indices and an
        up or down key press will move it to one of them.
        """
        self._min_line_length = wcswidth(PROMPT_1) + 1
        self.exec_in_thread = exec_in_thread
        self.alpha = alpha

        def fix_input_pos():
            self._input.left = self._prompt.right

        def update_cursor():
            self._input.cursor = self._input.cursor

        self._prompt.bind("size", fix_input_pos)
        self._prompt.bind("pos", fix_input_pos)
        self._scroll_view.bind("show_horizontal_bar", update_cursor)
        self._scroll_view.bind("show_vertical_bar", update_cursor)

        self._prompt.set_text(PROMPT_1)
        self._container.add_gadgets(self._output, self._prompt, self._input)

        self._scroll_view.view = self._container
        self.add_gadget(self._scroll_view)

        if banner is None:
            self._add_text_to_output(DEFAULT_BANNER)
        else:
            self._add_text_to_output(str(banner))

    def _add_text_to_output(self, text: str, with_prompt: bool = False):
        if with_prompt:
            text = f"{self._prompt._text}{text}"
        else:
            text = text.rstrip("\n")
            if not text:
                return

        lines = text.split("\n")
        max_line_length = max(wcswidth(line) for line in lines)

        if self._min_line_length < max_line_length:
            self._min_line_length = max_line_length

        if self._output.is_visible:
            line_number = self._output.height
            self._output.height += len(lines)
        else:
            self._output.is_visible = True
            self._output.y = 0
            line_number = 0
            self._output.height = len(lines)

        self._container.height += len(lines)
        self._container.width = self._min_line_length
        self._output.width = self._min_line_length
        for i, line in enumerate(lines):
            self._output.add_str(line, pos=(line_number + i, 0))
        self._prompt.top = self._output.bottom
        self._input.top = self._output.bottom

        self._scroll_view.horizontal_proportion = 0.0
        self._scroll_view.vertical_proportion = 1.0

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._scroll_view.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._scroll_view.alpha = alpha
        self._input.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._scroll_view.is_transparent = self.is_transparent
        self._input.is_transparent = self.is_transparent

    @property
    def exec_in_thread(self) -> bool:
        """
        Whether code is executed in a separate thread. Code not executed in a separated
        thread can block the event loop.
        """
        return self._console.exec_in_thread

    @exec_in_thread.setter
    def exec_in_thread(self, exec_in_thread: bool):
        self._console.exec_in_thread = exec_in_thread

    def update_theme(self):
        """Paint the gadget with current theme."""
        primary_fg = self.get_color("primary_fg")
        primary_bg = self.get_color("primary_bg")
        self._prompt.default_fg_color = self._prompt.canvas["fg_color"] = primary_fg
        self._prompt.default_bg_color = self._prompt.canvas["bg_color"] = primary_bg
        self._output.default_fg_color = self._output.canvas["fg_color"] = primary_fg
        self._output.default_bg_color = self._output.canvas["bg_color"] = primary_bg
        self._input._box.canvas["fg_color"] = primary_fg
        self._input._box.default_fg_color = primary_fg
        self._input._box.canvas["bg_color"] = primary_bg
        self._input._box.default_bg_color = primary_bg

    def on_add(self):
        """Add running app and root gadget to console's locals."""
        super().on_add()
        self._console.locals["app"] = self.app
        self._console.locals["root"] = self.root

    def on_focus(self):
        """Enable cursor."""
        self._input.focus()
        self._input._cursor.is_enabled = True

    def on_blur(self):
        """Disable cursor."""
        self._input._cursor.is_enabled = False

    def on_key(self, key_event: KeyEvent) -> bool | None:
        """Get previous inputs on up/down."""
        if (
            key_event.key == "up"
            and not key_event.alt
            and not key_event.ctrl
            and not key_event.shift
        ):
            if isinstance(self._history_index, float):
                if len(self._history) + self._history_index - 0.5 >= 0:
                    self._history_index = int(self._history_index - 0.5)
                    self._input.text = self._history[self._history_index]
            elif len(self._history) + self._history_index - 1 >= 0:
                self._history_index -= 1
                self._input.text = self._history[self._history_index]
        elif (
            key_event.key == "down"
            and not key_event.alt
            and not key_event.ctrl
            and not key_event.shift
        ):
            if isinstance(self._history_index, float):
                if self._history_index + 0.5 < 0:
                    self._history_index = int(self._history_index + 0.5)
                    self._input.text = self._history[self._history_index]
            elif self._history_index + 1 < 0:
                self._history_index += 1
                self._input.text = self._history[self._history_index]
        else:
            return super().on_key(key_event)

        return True



"""A scrollable view gadget."""

import asyncio
from typing import Final

from ..geometry import round_down
from ..terminal.events import KeyEvent, MouseButton, MouseEvent
from ..text_tools import Style, smooth_horizontal_bar, smooth_vertical_bar
from .behaviors.grabbable import Grabbable
from .behaviors.themable import Themable
from .gadget import Gadget, Point, PosHint, Size, SizeHint, bindable, clamp
from .pane import Pane
from .text import Text

__all__ = ["ScrollView", "Point", "Size"]

_INERTIAL_SCROLL_STRENGTH: Final = 2.4
"""Strength of inertial scroll."""
_DAMPING: Final = 0.95
"""Velocity damping for inertial scrolling."""
_VELOCITY_CUTOFF: Final = 0.001
"""Minimum velocity before stopping inertial scrolling."""


class _ScrollbarBase(Grabbable, Text):
    length: int

    def __init__(self):
        super().__init__(size=(1, 2))
        self.indicator_proportion: float = 1.0
        self.indicator_progress: float = 0.0
        self.is_hovered = False

    @property
    def indicator_proportion(self) -> float:
        return self._indicator_proportion

    @indicator_proportion.setter
    def indicator_proportion(self, indicator_porportion: float):
        self._indicator_proportion = indicator_porportion
        self._set_indicator_length()

    @property
    def fill_length(self) -> float:
        """The length the indicator can travel."""
        return self.length - self.indicator_length

    def _paint_indicator(self) -> None:
        sv: ScrollView = self.parent

        if self.is_grabbed:
            indicator_color = sv.get_color("scroll_view_indicator_press")
        elif self.is_hovered:
            indicator_color = sv.get_color("scroll_view_indicator_hover")
        else:
            indicator_color = sv.get_color("scroll_view_indicator_normal")

        self.chars[:] = " "
        self.canvas["fg_color"] = indicator_color
        self.canvas["bg_color"] = sv.get_color("scroll_view_scrollbar")
        self.canvas["style"] = 0

        start, offset = divmod(self.indicator_progress * self.fill_length, 1)
        start = int(start)
        # Round offset to the nearest 1/8th.
        offset = round_down(offset * 8) / 8
        if offset == 1:
            offset -= 1
            start += 1

        self._start = start
        self._offset = offset

    def ungrab(self, mouse_event):
        super().ungrab(mouse_event)
        self._paint_indicator()


class _VerticalScrollbar(_ScrollbarBase):
    @property
    def length(self) -> int:
        return self.height

    def _set_indicator_length(self):
        self.indicator_length = clamp(
            2, round_down(self.indicator_proportion * self.length), self.length
        )

    def _paint_indicator(self) -> None:
        super()._paint_indicator()
        if self.indicator_length == 0:
            return
        start, offset = self._start, self._offset
        smooth_bar = smooth_vertical_bar(
            self.indicator_length, 1, offset, reversed=True
        )
        stop = start + len(smooth_bar)
        self.chars[start:stop].T[:] = smooth_bar
        self.canvas["style"][start + bool(offset) : stop] = Style.REVERSE

    def on_mouse(self, mouse_event):
        old_hovered = self.is_hovered

        y, x = self.to_local(mouse_event.pos)
        self.is_hovered = 0 <= x < 2 and (
            self._start <= y < self._start + self.indicator_length + bool(self._offset)
        )

        if super().on_mouse(mouse_event):
            return True

        if old_hovered != self.is_hovered:
            self._paint_indicator()

    def grab(self, mouse_event):
        super().grab(mouse_event)

        sv: ScrollView = self.parent

        if sv._inertial_scroll_task is not None:
            sv._inertial_scroll_task.cancel()

        if self.is_hovered:
            self._paint_indicator()
        else:
            self.is_hovered = True

            if self.fill_length == 0:
                sv.vertical_proportion = 0
            else:
                sv.vertical_proportion = self.to_local(mouse_event.pos).y / self.length

    def grab_update(self, mouse_event):
        sv: ScrollView = self.parent
        if self.fill_length == 0:
            sv.vertical_proportion = 0
        else:
            sv.vertical_proportion += mouse_event.dy / self.fill_length


class _HorizontalScrollbar(_ScrollbarBase):
    @property
    def length(self):
        return self.width

    def _set_indicator_length(self):
        self.indicator_length = clamp(
            4, round_down(self.indicator_proportion * self.length), self.length
        )

    def _paint_indicator(self) -> None:
        super()._paint_indicator()
        if self.indicator_length == 0:
            return
        start, offset = self._start, self._offset
        smooth_bar = smooth_horizontal_bar(self.indicator_length, 1, offset)
        self.chars[:, start : start + len(smooth_bar)] = smooth_bar
        if offset:
            self.canvas["style"][:, start] = Style.REVERSE

    def on_mouse(self, mouse_event):
        old_hovered = self.is_hovered

        y, x = self.to_local(mouse_event.pos)
        self.is_hovered = y == 0 and (
            self._start <= x < self._start + self.indicator_length + bool(self._offset)
        )

        if super().on_mouse(mouse_event):
            return True

        if old_hovered != self.is_hovered:
            self._paint_indicator()

    def grab(self, mouse_event):
        super().grab(mouse_event)

        sv: ScrollView = self.parent

        if sv._inertial_scroll_task is not None:
            sv._inertial_scroll_task.cancel()

        if self.is_hovered:
            self._paint_indicator()
        else:
            self.is_hovered = True

            if self.fill_length == 0:
                sv.horizontal_proportion = 0
            else:
                sv.horizontal_proportion = (
                    self.to_local(mouse_event.pos).x / self.length
                )

    def grab_update(self, mouse_event):
        sv: ScrollView = self.parent
        if self.fill_length == 0:
            sv.horizontal_proportion = 0
        else:
            sv.horizontal_proportion += mouse_event.dx / self.fill_length


class ScrollView(Themable, Grabbable, Gadget):
    r"""
    A scrollable view gadget.

    The view can be set with the :attr:`view` property, e.g.,
    ``my_scroll_view.view = some_gadget``.

    Parameters
    ----------
    allow_vertical_scroll : bool, default: True
        Allow vertical scrolling.
    allow_horizontal_scroll : bool, default: True
        Allow horizontal scrolling.
    show_vertical_bar : bool, default: True
        Whether the vertical scrollbar is shown.
    show_horizontal_bar : bool, default: True
        Whether the horizontal scrollbar is shown.
    dynamic_bars : bool, default: False
        Whether bars are shown/hidden depending on the view size.
    scrollwheel_enabled : bool, default: True
        Allow vertical scrolling with scrollwheel.
    arrow_keys_enabled : bool, default: True
        Allow scrolling with arrow keys.
    inertial_scrolling_enabled : bool, default: False
        Whether inertial scrolling is enabled.
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton, default: "left"
        Mouse button used for grabbing.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    view : Gadget | None
        The scrolled gadget.
    allow_vertical_scroll : bool
        Allow vertical scrolling.
    allow_horizontal_scroll : bool
        Allow horizontal scrolling.
    show_vertical_bar : bool
        Whether the vertical scrollbar is shown.
    show_horizontal_bar : bool
        Whether the horizontal scrollbar is shown.
    dynamic_bars : bool, default: False
        Whether bars are shown/hidden depending on the view size.
    scrollwheel_enabled : bool
        Allow vertical scrolling with scrollwheel.
    arrow_keys_enabled : bool
        Allow scrolling with arrow keys.
    inertial_scrolling_enabled : bool
        Whether inertial scrolling is enabled.
    vertical_proportion : float
        Vertical scroll position as a proportion of total height.
    horizontal_proportion : float
        Horizontal scroll position as a proportion of total width.
    port_height : int
        Height of view port.
    port_width : int
        Width of view port.
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton
        Mouse button used for grabbing.
    is_grabbed : bool
        Whether gadget is grabbed.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    scroll_left(n=1)
        Scroll the view left `n` characters.
    scroll_right(n=1)
        Scroll the view right `n` characters.
    scroll_up(n=1)
        Scroll the view up `n` lines.
    scroll_down(n=1)
        Scroll the view down `n` lines.
    scroll_to_rect(pos, size=(1, 1))
        Scroll the view so that a given rect is visible.
    get_color()
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    grab(mouse_event)
        Grab the gadget.
    ungrab(mouse_event)
        Ungrab the gadget.
    grab_update(mouse_event)
        Update gadget with incoming mouse events while grabbed.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        allow_vertical_scroll: bool = True,
        allow_horizontal_scroll: bool = True,
        show_vertical_bar: bool = True,
        show_horizontal_bar: bool = True,
        dynamic_bars: bool = False,
        scrollwheel_enabled: bool = True,
        arrow_keys_enabled: bool = True,
        inertial_scrolling_enabled: bool = False,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton = "left",
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._background = Pane(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
            alpha=alpha,
            is_transparent=is_transparent,
        )
        self._vertical_bar = _VerticalScrollbar()
        self._horizontal_bar = _HorizontalScrollbar()
        self._corner = Pane(
            size=(1, 2),
            pos_hint={"y_hint": 1.0, "x_hint": 1.0, "anchor": "bottom-right"},
            is_enabled=show_horizontal_bar or show_vertical_bar,
            is_transparent=False,
        )
        self._vertical_proportion = 0.0
        self._horizontal_proportion = 0.0
        self._view = None
        self._vertical_velocity: float = 0.0
        """Vertical velocity for inertial scrolling."""
        self._horizontal_velocity: float = 0.0
        """Horizontal velocity for inertial scrolling."""
        self._inertial_scroll_task: asyncio.Task | None = None
        """Task that updates inertial scroll."""

        super().__init__(
            is_grabbable=is_grabbable,
            ptf_on_grab=ptf_on_grab,
            mouse_button=mouse_button,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.allow_vertical_scroll: bool = allow_vertical_scroll
        """Allow vertical scrolling."""
        self.allow_horizontal_scroll: bool = allow_horizontal_scroll
        """Allow horizontal scrolling."""
        self.inertial_scrolling_enabled: bool = inertial_scrolling_enabled
        """Whether inertial scrolling is enabled."""
        self.scrollwheel_enabled = scrollwheel_enabled
        """Allow vertical scrolling with scrollwheel."""
        self.arrow_keys_enabled = arrow_keys_enabled
        """Allow scrolling with arrow keys."""

        self.add_gadgets(
            self._background, self._corner, self._vertical_bar, self._horizontal_bar
        )

        if not dynamic_bars:
            self._vertical_bar.is_enabled = show_vertical_bar
            self._horizontal_bar.is_enabled = show_horizontal_bar
            self._corner.is_enabled = show_horizontal_bar or show_vertical_bar
        self.dynamic_bars = dynamic_bars

    def update_theme(self):
        """Paint the gadget with current theme."""
        self._background.bg_color = self.get_color("primary_bg")
        self._corner.bg_color = self.get_color("scroll_view_scrollbar")
        self._update_port_and_scrollbar()

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._background.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._background.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._background.is_transparent = self.is_transparent

    @property
    def show_vertical_bar(self) -> bool:
        """Whether the vertical scrollbar is shown."""
        return self._vertical_bar.is_enabled

    @show_vertical_bar.setter
    @bindable
    def show_vertical_bar(self, show: bool):
        self._vertical_bar.is_enabled = show
        self._corner.is_enabled = show or self._horizontal_bar.is_enabled
        self.on_size()

    @property
    def show_horizontal_bar(self) -> bool:
        """Whether the horizontal scrollbar is shown."""
        return self._horizontal_bar.is_enabled

    @show_horizontal_bar.setter
    @bindable
    def show_horizontal_bar(self, show: bool):
        self._horizontal_bar.is_enabled = show
        self._corner.is_enabled = show or self._vertical_bar.is_enabled
        self.on_size()

    @property
    def dynamic_bars(self) -> bool:
        """Whether bars are shown/hidden depending on the view size."""
        return self._dynamic_bars

    @dynamic_bars.setter
    def dynamic_bars(self, dynamic_bars: bool):
        self._dynamic_bars = dynamic_bars
        self.on_size()

    @property
    def vertical_proportion(self) -> float:
        """Vertical scroll position as a proportion of total height."""
        return self._vertical_proportion

    @vertical_proportion.setter
    @bindable
    def vertical_proportion(self, vertical_proportion: float):
        if self.allow_vertical_scroll:
            if self._view is None:
                self._vertical_proportion = 0.0
            else:
                self._vertical_proportion = clamp(float(vertical_proportion), 0.0, 1.0)
            self._update_port_and_scrollbar()

    @property
    def horizontal_proportion(self) -> float:
        """Horizontal scroll position as a proportion of total width."""
        return self._horizontal_proportion

    @horizontal_proportion.setter
    @bindable
    def horizontal_proportion(self, horizontal_proportion: float):
        if self.allow_horizontal_scroll:
            if self._view is None:
                self._horizontal_proportion = 0.0
            else:
                self._horizontal_proportion = clamp(
                    float(horizontal_proportion), 0.0, 1.0
                )
            self._update_port_and_scrollbar()

    @property
    def port_height(self) -> int:
        """Height of view port."""
        return max(0, self.height - self.show_horizontal_bar)

    @property
    def port_width(self) -> int:
        """Width of view port."""
        return max(0, self.width - self.show_vertical_bar * 2)

    @property
    def total_vertical_distance(self) -> int:
        """The distance the view can scroll vertically."""
        return 0 if self._view is None else max(0, self._view.height - self.port_height)

    @property
    def total_horizontal_distance(self) -> int:
        """The distance the view can scroll horizontally."""
        return 0 if self._view is None else max(0, self._view.width - self.port_width)

    def _update_port_and_scrollbar(self):
        """Move port and repaint scrollbar."""
        if self._view is None:
            self._vertical_bar.indicator_proportion = 1.0
            self._vertical_bar.indicator_progress = 0
            self._vertical_bar._paint_indicator()
            self._horizontal_bar.indicator_proportion = 1.0
            self._horizontal_bar.indicator_progress = 0
            self._horizontal_bar._paint_indicator()
        else:
            self._view.top = -round_down(
                self.vertical_proportion * self.total_vertical_distance
            )
            self._view.left = -round_down(
                self.horizontal_proportion * self.total_horizontal_distance
            )

            self._vertical_bar.indicator_proportion = clamp(
                self.port_height / self._view.height, 0, 1
            )
            self._vertical_bar.indicator_progress = self.vertical_proportion
            self._vertical_bar._paint_indicator()

            self._horizontal_bar.indicator_proportion = clamp(
                self.port_width / self._view.width, 0, 1
            )
            self._horizontal_bar.indicator_progress = self.horizontal_proportion
            self._horizontal_bar._paint_indicator()

    @property
    def view(self) -> Gadget | None:
        """The scrolled gadget."""
        return self._view

    @view.setter
    def view(self, view: Gadget | None):
        if self._view is not None:
            self.remove_gadget(self._view)

        self._view = view
        if view is not None:
            self.add_gadget(view)
            self.children.insert(1, self.children.pop())  # Move view below scrollbars.

            def update_proportion():
                y, x = self._view.pos
                h = self.total_vertical_distance
                w = self.total_horizontal_distance
                self._vertical_proportion = 0.0 if h == 0 else clamp(-y / h, 0.0, 1.0)
                self._horizontal_proportion = 0.0 if w == 0 else clamp(-x / w, 0.0, 1.0)
                self.on_size()

            view.pos = Point(0, 0)
            update_proportion()
            self._view_bind_uid = view.bind("size", update_proportion)

    def _inertial_scroll(self):
        if not self.inertial_scrolling_enabled:
            return

        if self._inertial_scroll_task is not None:
            self._inertial_scroll_task.cancel()
        self._inertial_scroll_task = asyncio.create_task(self._damp_velocity())

    async def _damp_velocity(self):
        while True:
            self._vertical_velocity *= _DAMPING
            if abs(self._vertical_velocity) <= _VELOCITY_CUTOFF:
                self._vertical_velocity = 0
            else:
                self.vertical_proportion += self._vertical_velocity

            self._horizontal_velocity *= _DAMPING
            if abs(self._horizontal_velocity) <= _VELOCITY_CUTOFF:
                self._horizontal_velocity = 0
            else:
                self.horizontal_proportion += self._horizontal_velocity

            if self._vertical_velocity == 0 and self._horizontal_velocity == 0:
                return

            try:
                await asyncio.sleep(0.1)
            except asyncio.CancelledError as e:
                self._vertical_velocity = 0
                self._horizontal_velocity = 0
                raise asyncio.CancelledError from e

    def on_remove(self) -> None:
        """Cancel inertial scroll on remove."""
        if self._inertial_scroll_task is not None:
            self._inertial_scroll_task.cancel()
            self._inertial_scroll_task = None
        super().on_remove()

    def remove_gadget(self, gadget: Gadget):
        """Unbind from the view on its removal."""
        if gadget is self._view:
            self._view.unbind(self._view_bind_uid)
            del self._view_bind_uid
            self._view = None

        super().remove_gadget(gadget)

    def on_size(self):
        """Resize and reposition scrollbars on resize."""
        if self.dynamic_bars:
            if self.view is None:
                self._vertical_bar.is_enabled = False
                self._horizontal_bar.is_enabled = False
                self._corner.is_enabled = False
            else:
                vh, vw = self.view.size
                h, w = self.size
                if h < vh:
                    self._vertical_bar.is_enabled = True
                    self._horizontal_bar.is_enabled = w < vw + 2
                    self._corner.is_enabled = True
                elif w < vw:
                    self._horizontal_bar.is_enabled = True
                    self._vertical_bar.is_enabled = h < vh + 1
                    self._corner.is_enabled = True
                else:
                    self._vertical_bar.is_enabled = False
                    self._horizontal_bar.is_enabled = False
                    self._corner.is_enabled = False

        self._vertical_bar.height = self.height - self.show_horizontal_bar
        self._vertical_bar.left = self.width - 2
        self._horizontal_bar.width = self.width - 2 * self.show_vertical_bar
        self._horizontal_bar.top = self.height - 1
        self._update_port_and_scrollbar()

    def on_key(self, key_event: KeyEvent) -> bool | None:
        """Scroll on arrows keys."""
        if not self.arrow_keys_enabled:
            return False

        if key_event.key == "up":
            self.scroll_up()
        elif key_event.key == "down":
            self.scroll_down()
        elif key_event.key == "left":
            self.scroll_left()
        elif key_event.key == "right":
            self.scroll_right()
        else:
            return super().on_key(key_event)

        return True

    def grab(self, mouse_event: MouseEvent) -> None:
        """Cancel inertial scroll."""
        super().grab(mouse_event)
        if self._inertial_scroll_task is not None:
            self._inertial_scroll_task.cancel()

    def grab_update(self, mouse_event: MouseEvent) -> None:
        """Scroll on grab update."""
        if self.view is None:
            return

        self.scroll_up(mouse_event.dy)
        self.scroll_left(mouse_event.dx)
        if self.view.height:
            vv = -_INERTIAL_SCROLL_STRENGTH * mouse_event.dy / self.view.height
        else:
            vv = 0
        if abs(vv) > abs(self._vertical_velocity) or vv * self._vertical_velocity < 0:
            self._vertical_velocity = vv

        if self.view.width:
            hv = -_INERTIAL_SCROLL_STRENGTH * mouse_event.dx / self.view.width
        else:
            hv = 0
        if (
            abs(hv) > abs(self._horizontal_velocity)
            or hv * self._horizontal_velocity < 0
        ):
            self._horizontal_velocity = hv

    def ungrab(self, mouse_event: MouseEvent) -> None:
        """Start inertial scroll."""
        super().ungrab(mouse_event)
        self._inertial_scroll()

    def on_mouse(self, mouse_event: MouseEvent) -> bool | None:
        """Scroll on mouse wheel."""
        if self.scrollwheel_enabled and self.collides_point(mouse_event.pos):
            if mouse_event.event_type == "scroll_up":
                self.scroll_up()
            elif mouse_event.event_type == "scroll_down":
                self.scroll_down()
            else:
                return super().on_mouse(mouse_event)

            if self._inertial_scroll_task is not None:
                self._inertial_scroll_task.cancel()
            return True

        return super().on_mouse(mouse_event)

    def scroll_left(self, n=1):
        """Scroll the view left `n` characters."""
        if self._view is not None:
            if self.total_horizontal_distance == 0:
                self.horizontal_proportion = 0
            else:
                self.horizontal_proportion = clamp(
                    (-self.view.left - n) / self.total_horizontal_distance, 0, 1
                )

    def scroll_right(self, n=1):
        """Scroll the view right `n` characters."""
        self.scroll_left(-n)

    def scroll_up(self, n=1):
        """Scroll the view up `n` lines."""
        if self._view is not None:
            if self.total_vertical_distance == 0:
                self.vertical_proportion = 0
            else:
                self.vertical_proportion = clamp(
                    (-self.view.top - n) / self.total_vertical_distance, 0, 1
                )

    def scroll_down(self, n=1):
        """Scroll the view down `n` lines."""
        self.scroll_up(-n)

    def scroll_to_rect(self, pos: Point, size: Size = Size(1, 1)):
        """
        Scroll the view so that a given rect is visible.

        Does nothing if the scroll view or one it's scrollbars are grabbed.

        Parameters
        ----------
        pos : Point
            Position of rect.

        size : Size, default: Size(1, 1)
            Size of rect.
        """
        if (
            self.view is None
            or self.is_grabbed
            or self._horizontal_bar.is_grabbed
            or self._vertical_bar.is_grabbed
        ):
            return

        if self._inertial_scroll_task is not None:
            self._inertial_scroll_task.cancel()

        h, w = size
        ay, ax = self.view.pos + pos
        if ay < 0:
            self.scroll_up(-ay)
        elif h >= self.port_height:
            self.scroll_down(ay)
        elif h + ay >= self.port_height:
            self.scroll_down(ay + h - self.port_height)

        if ax < 0:
            self.scroll_left(-ax)
        elif w >= self.port_width:
            self.scroll_right(ax)
        elif w + ax >= self.port_width:
            self.scroll_right(ax + w - self.port_width)



"""Platform specific VT100 terminals."""

import asyncio
import sys
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from typing import Final

from ..geometry import Size
from .events import (
    DECReplyModeEvent,
    DeviceAttributesReportEvent,
    Event,
    PixelGeometryReportEvent,
)
from .vt100_terminal import DSR_REQUEST_TIMEOUT, Vt100Terminal

__all__ = [
    "Vt100Terminal",
    "app_mode",
    "determine_terminal_capabilities",
    "get_platform_terminal",
]

_SIXEL_SUPPORT: Final = 4
"""Terminal attribute for sixel support."""


@contextmanager
def app_mode(
    terminal: Vt100Terminal, event_handler: Callable[[list[Event]], None]
) -> Iterator[None]:
    """
    Put terminal into app mode and dispatch input events.

    Parameters
    ----------
    terminal : Vt100Terminal
        Terminal to put in app mode.
    event_handler : Callable[[list[Event]], None]
        A callable that handles terminal input events.
    """
    try:
        terminal.raw_mode()
        terminal.attach(event_handler)
        terminal.enable_mouse_support()
        terminal.enable_bracketed_paste()
        terminal.enable_reporting_focus()
        terminal.hide_cursor()
        terminal.flush()
        yield
    finally:
        if terminal.in_alternate_screen:
            terminal.exit_alternate_screen()
        else:
            terminal.erase_in_display()
        terminal.reset_attributes()
        terminal.show_cursor()
        terminal.disable_reporting_focus()
        terminal.disable_bracketed_paste()
        terminal.disable_mouse_support()
        terminal.flush()
        terminal.unattach()
        terminal.restore_console()


def get_platform_terminal() -> Vt100Terminal:
    """
    Return a platform-specific terminal.

    Returns
    -------
    Vt100Terminal
        A platform-specific VT100 terminal.

    Raises
    ------
    RuntimeError
        If terminal isn't interactive or terminal doesn't support VT100 sequences.
    """
    if sys.platform == "win32":
        from .windows_terminal import WindowsTerminal

        return WindowsTerminal()
    else:
        from .linux_terminal import LinuxTerminal

        return LinuxTerminal()


async def determine_terminal_capabilities(terminal: Vt100Terminal) -> tuple[bool, Size]:
    """
    Determine various terminal capabilities.

    Try to determine terminal's pixel geometry. If pixel geometry is determined,
    determine if terminal has sixel support and attempt to turn on SGR-PIXELS mouse
    mode. Additionally, query for synchronized update mode support.

    Parameters
    ----------
    terminal : VT100Terminal
        The VT100 terminal to query.

    Returns
    -------
    tuple[bool, Size]
        Return whether sixel is supported and pixel geometry.
    """
    pixel_geometry: Size | None = None
    sixel_support: bool = False
    terminal_info_reported: asyncio.Event = asyncio.Event()
    report_timeout: asyncio.TimerHandle

    def report_handler(events: list[Event]) -> None:
        """Handle terminal reports."""
        nonlocal sixel_support, pixel_geometry, report_timeout

        for event in events:
            if isinstance(event, DeviceAttributesReportEvent):
                sixel_support = _SIXEL_SUPPORT in event.device_attributes
            elif isinstance(event, PixelGeometryReportEvent):
                if event.kind == "cell":
                    pixel_geometry = event.geometry
                else:
                    ph, pw = event.geometry
                    th, tw = terminal.get_size()
                    pixel_geometry = Size(ph // th, pw // tw)
            elif isinstance(event, DECReplyModeEvent):
                if event.mode != 1016:
                    continue

                if event.value:
                    terminal.enable_sgr_pixels()
            else:
                continue

            report_timeout.cancel()
            terminal_info_reported.set()

    async def wait_for_report():
        nonlocal report_timeout
        report_timeout = asyncio.get_running_loop().call_later(
            DSR_REQUEST_TIMEOUT, terminal_info_reported.set
        )
        await terminal_info_reported.wait()
        terminal_info_reported.clear()

    old_handler = terminal._event_handler
    terminal._event_handler = report_handler

    terminal.request_synchronized_update_mode_supported()
    # Result of request is stored in `terminal`.

    terminal.request_pixel_geometry()
    await wait_for_report()

    if pixel_geometry is None:
        # Fallback to terminal geometry from which pixel geometry can be calculated.
        terminal.request_terminal_geometry()
        await wait_for_report()

    if pixel_geometry is None:
        # No pixel geometry reported; don't even attempt SGR-PIXELS or SIXEL.
        return False, Size(20, 10)

    terminal.request_sgr_pixels_supported()
    await wait_for_report()

    terminal.request_device_attributes()
    await wait_for_report()

    terminal._event_handler = old_handler
    return sixel_support, pixel_geometry



"""A text-pad gadget for multiline editable text."""

from dataclasses import astuple

from ugrapheme import grapheme_iter, graphemes
from uwcwidth import wcswidth

from ..logging import get_logger
from ..terminal.events import KeyEvent, MouseEvent, PasteEvent
from ..text_tools import canvas_as_text, egc_chr, is_word_char
from .behaviors.focusable import Focusable
from .behaviors.grabbable import Grabbable
from .behaviors.themable import Themable
from .cursor import Cursor
from .gadget import Gadget, Point, PosHint, Size, SizeHint
from .scroll_view import ScrollView
from .text import Text

__all__ = ["TextPad", "Point", "Size"]

logger = get_logger(__name__)


class TextPad(Themable, Grabbable, Focusable, Gadget):
    r"""
    A text-pad gadget for multiline editable text.

    Supports pasting, mouse selection, and cursor navigation.

    Parameters
    ----------
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    alpha : float
        Transparency of gadget.
    text : str
        The text pad's text.
    cursor : Point
        The cursor position.
    is_selecting : bool
        Whether there is a selection.
    has_nonempty_selection : bool
        Whether selection is non-empty.
    page_lines : int
        Number of rows a page-up or -down moves.
    is_focused : bool
        Whether gadget has focus.
    any_focused : bool
        Whether any gadget has focus.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    undo()
        Undo previous edit.
    redo()
        Redo previous undo.
    select()
        Start a new selection at cursor if none.
    unselect()
        Unselect current selection.
    delete_selection()
        Delete current selection.
    move_cursor_left(n)
        Move cursor left `n` characters.
    move_cursor_right(n)
        Move cursor right `n` characters.
    move_cursor_up(n)
        Move cursor up `n` rows.
    move_cursor_down(n)
        Move cursor down `n` rows.
    move_word_left()
        Move cursor a word left.
    move_word_right()
        Move cursor a word right.
    get_color()
        Get a color by name from the current color theme.
    update_theme()
        Paint the gadget with current theme.
    focus()
        Focus gadget.
    blur()
        Un-focus gadget.
    focus_next()
        Focus next focusable gadget.
    focus_previous()
        Focus previous focusable gadget.
    on_focus()
        Update gadget when it gains focus.
    on_blur()
        Update gadget when it loses focus.
    grab(mouse_event)
        Grab the gadget.
    ungrab(mouse_event)
        Ungrab the gadget.
    grab_update(mouse_event)
        Update gadget with incoming mouse events while grabbed.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._cursor = Cursor()
        self._pad = Text(size=(1, 1), is_transparent=is_transparent)
        self._scroll_view = ScrollView(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
            arrow_keys_enabled=False,
            is_grabbable=False,
            alpha=0,
            is_transparent=is_transparent,
        )
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self._last_x = None
        self._selection_start = self._selection_end = None
        self._line_widths = [0]
        self._undo_stack = []
        self._redo_stack = []
        self._undo_buffer = []
        self._undo_buffer_type = "add"
        self.alpha = alpha

        self._pad.add_gadget(self._cursor)
        self._scroll_view.view = self._pad
        self.add_gadget(self._scroll_view)

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._pad.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._pad.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._pad.is_transparent = self.is_transparent
        self._scroll_view.is_transparent = self.is_transparent

    def update_theme(self):
        """Paint the gadget with current theme."""
        primary_fg = self.get_color("primary_fg")
        primary_bg = self.get_color("primary_bg")

        self._pad.canvas["fg_color"] = self._pad.default_fg_color = primary_fg
        self._pad.canvas["bg_color"] = self._pad.default_bg_color = primary_bg
        self._highlight_selection()

    def on_add(self):
        """Bind pad resize to scroll view resize."""
        super().on_add()

        def resize_pad():
            height = max(len(self._line_widths), self._scroll_view.port_height)
            width = max(max(self._line_widths) + 1, self._scroll_view.port_width)
            self._pad.size = height, width
            self._highlight_selection()

        resize_pad()
        self._bind_uid = self._scroll_view.bind("size", resize_pad)

    def on_remove(self):
        """Unbind pad resize from scroll view resize."""
        self._scroll_view.unbind(self._bind_uid)
        super().on_remove()

    def on_focus(self):
        """Show cursor on focus."""
        self._cursor.is_enabled = True

    def on_blur(self):
        """Hide cursor on blur."""
        self._cursor.is_enabled = False

    def _move_undo_buffer_to_stack(self, buffer_type=None):
        self._undo_buffer_type = buffer_type
        if self._undo_buffer:
            self._undo_stack.append(self._undo_buffer)
            self._undo_buffer = []
            self._redo_stack.clear()

    def undo(self):
        """Undo previous edit."""
        self._move_undo_buffer_to_stack()
        if self._undo_stack:
            redo = []
            for func, args, selection_start, selection_end, cursor in reversed(
                self._undo_stack.pop()
            ):
                redo.append(func(*args))
                self._selection_start = selection_start
                self._selection_end = selection_end
                self.cursor = cursor
            self._redo_stack.append(redo)

    def redo(self):
        """Redo previous undo."""
        if self._redo_stack and not self._undo_buffer:
            undo = []
            for func, args, selection_start, selection_end, cursor in reversed(
                self._redo_stack.pop()
            ):
                undo.append(func(*args))
                self._selection_start = selection_start
                self._selection_end = selection_end
                self.cursor = cursor
            self._undo_stack.append(undo)

    @property
    def text(self) -> str:
        """The text pad's text."""
        return canvas_as_text(self._pad.canvas, self._line_widths)

    @text.setter
    def text(self, text: str):
        self.unselect()
        self._del_text((0, 0), self.end_text_point)
        self._add_text((0, 0), text)
        self._redo_stack.clear()
        self._undo_stack.clear()
        self._undo_buffer.clear()
        self.cursor = self.end_text_point

    @property
    def cursor(self) -> Point:
        """The cursor position."""
        return self._cursor.pos

    @cursor.setter
    def cursor(self, cursor: Point):
        """After setting cursor position, move pad so that cursor is visible."""
        self._cursor.pos = cursor
        self._scroll_view.scroll_to_rect(cursor)
        if self.is_selecting:
            self._selection_end = self.cursor
        self._highlight_selection()

    def _highlight_selection(self):
        fg = self._pad.canvas["fg_color"]
        bg = self._pad.canvas["bg_color"]
        fg[:] = self._pad.default_fg_color
        bg[:] = self._pad.default_bg_color

        if self._selection_start != self._selection_end:
            if self._selection_start > self._selection_end:
                sy, sx = self._selection_end
                ey, ex = self._selection_start
            else:
                sy, sx = self._selection_start
                ey, ex = self._selection_end

            highlight_fg = self.get_color("text_pad_selection_highlight_fg")
            highlight_bg = self.get_color("text_pad_selection_highlight_bg")
            ll = self._line_widths
            if sy == ey:
                fg[sy, sx:ex] = highlight_fg
                bg[sy, sx:ex] = highlight_bg
            else:
                fg[sy, sx : ll[sy]] = highlight_fg
                bg[sy, sx : ll[sy]] = highlight_bg
                fg[ey, :ex] = highlight_fg
                bg[ey, :ex] = highlight_bg
                for i in range(sy + 1, ey):
                    fg[i, : ll[i]] = highlight_fg
                    bg[i, : ll[i]] = highlight_bg
        else:  # If no selection or selection is empty, add line highlight.
            fg[self.cursor.y, :] = self.get_color("text_pad_line_highlight_fg")
            bg[self.cursor.y, :] = self.get_color("text_pad_line_highlight_bg")

    @property
    def is_selecting(self) -> bool:
        """Whether there is a selection."""
        return self._selection_start is not None and self._selection_end is not None

    @property
    def has_nonempty_selection(self) -> bool:
        """Whether selection is non-empty."""
        return self.is_selecting and self._selection_start != self._selection_end

    @property
    def end_text_point(self) -> Point:
        """Point after last character in text."""
        ll = self._line_widths
        return Point(len(ll) - 1, ll[-1])

    @property
    def page_lines(self) -> int:
        """Number of rows a page-up or -down moves."""
        return self._scroll_view.port_height

    def select(self):
        """Start a new selection at cursor if none."""
        if not self.is_selecting:
            self._selection_start = self._selection_end = self.cursor

    def unselect(self):
        """Unselect current selection."""
        self._selection_start = self._selection_end = None

    def delete_selection(self):
        """Delete current selection."""
        if self.has_nonempty_selection:
            return self._del_text(self._selection_start, self._selection_end)

    def _del_text(self, start: Point, end: Point):
        ll = self._line_widths
        pad = self._pad

        if start > end:
            start, end = end, start

        sy, sx = start
        ey, ex = end

        if ey >= len(ll):
            logger.debug("End y-coordinate is greater than number of lines.")
            ey = len(ll) - 1
        if ex > ll[ey]:
            logger.debug("End x-coordinate is greater than line length.")
            ex = ll[ey]

        if sy == ey:
            contents = canvas_as_text(pad.canvas[sy, sx:ex])
        else:
            lines = [canvas_as_text(pad.canvas[sy, sx : ll[sy]])]
            if ey - sy > 1:
                lines.append(canvas_as_text(pad.canvas[sy + 1 : ey], ll[sy + 1 : ey]))
            lines.append(canvas_as_text(pad.canvas[ey, : ll[ey]]))
            contents = "\n".join(lines)

        selection_start = self._selection_start
        selection_end = self._selection_end
        cursor = self.cursor

        len_end = ll[ey] - ex
        len_start = ll[sy] = sx + len_end
        if len_start + 1 > pad.width:
            pad.width = len_start + 1

        pad.canvas[sy, sx:len_start] = pad.canvas[ey, ex : ex + len_end]
        pad.canvas[sy, len_start:] = pad.default_cell

        remaining = pad.canvas[ey + 1 :]
        pad.canvas[sy + 1 : sy + 1 + len(remaining)] = remaining
        pad.canvas[sy + 1 + len(remaining) :] = pad.default_cell

        del ll[sy + 1 : ey + 1]
        height = max(len(ll), self._scroll_view.port_height)
        width = max(max(ll) + 1, self._scroll_view.port_width)
        pad.size = height, width

        self.unselect()
        self._last_x = None
        self.cursor = start
        return self._add_text, [start, contents], selection_start, selection_end, cursor

    def _add_text(self, pos: Point, text: str):
        y, x = pos
        pad = self._pad
        ll = self._line_widths
        line_remaining = pad.canvas[y, x : ll[y]].copy()

        selection_start = self._selection_start
        selection_end = self._selection_end
        cursor = self.cursor

        lines = text.split("\n")  # DO NOT USE `splitlines`.
        if len(lines) == 1:
            line = lines[0]
            width_line = wcswidth(line)

            ll[y] += width_line
            if ll[y] >= pad.width:
                pad.width = ll[y] + 1

            pad.add_str(line, pos=(y, x))
            pad.canvas[y, x + width_line : ll[y]] = line_remaining
            self.cursor = y, x + width_line
        else:
            first, *lines, last = lines
            newlines = len(lines) + 1
            width_last = wcswidth(last)
            last_y = y + newlines

            ll[y] = x + wcswidth(first)
            for i, line in enumerate(lines, start=y + 1):
                ll.insert(i, wcswidth(line))
            ll.insert(last_y, width_last + wcswidth(canvas_as_text(line_remaining)))

            height = max(len(ll), self._scroll_view.port_height)
            width = max(max(ll) + 1, self._scroll_view.port_width)
            pad.size = height, width

            pad.canvas[y + newlines + 1 :] = pad.canvas[y + 1 : -newlines]
            pad.canvas[y, ll[y] :] = pad.default_cell

            pad.add_str(first, pos=(y, x))
            for i, line in enumerate(lines, start=y + 1):
                pad.add_str(line.ljust(pad.width), pos=(i, 0))

            pad.add_str(last, pos=(last_y, 0))
            pad.canvas[last_y, width_last : ll[last_y]] = line_remaining
            pad.canvas[last_y, ll[last_y] :] = pad.default_cell

            self.cursor = last_y, width_last

        return (
            self._del_text,
            [cursor, self.cursor],
            selection_start,
            selection_end,
            cursor,
        )

    def move_cursor_left(self, n: int = 1):
        """Move cursor left `n` characters."""
        self._last_x = None
        y, x = self._cursor.pos

        while n > 0:
            text_before_cursor = canvas_as_text(self._pad.canvas[y, :x])
            egcs = graphemes(text_before_cursor)
            if n <= len(egcs):
                x = wcswidth(egcs[:-n])
                break

            if y == 0:
                x = 0
                break

            y -= 1
            x = self._line_widths[y]
            n -= len(egcs) + 1

        self.cursor = y, x

    def move_cursor_right(self, n: int = 1):
        """Move cursor right `n` characters."""
        self._last_x = None
        y, x = self._cursor.pos

        while n > 0:
            text_after_cursor = canvas_as_text(
                self._pad.canvas[y, x : self._line_widths[y]]
            )
            egcs = graphemes(text_after_cursor)
            if n <= len(egcs):
                x += wcswidth(egcs[:n])
                break

            if y == self.end_text_point.y:
                x = self._line_widths[y]
                break

            y += 1
            n -= len(egcs) + 1
            x = 0

        self.cursor = y, x

    def _fix_x(self, y, x):
        line = canvas_as_text(self._pad.canvas[y, : self._line_widths[y]])
        current_x = 0
        for egc in grapheme_iter(line):
            if current_x + wcswidth(egc) > x:
                return current_x
            current_x += wcswidth(egc)
        return self._line_widths[y]

    def move_cursor_up(self, n: int = 1):
        """Move cursor up `n` rows."""
        y, x = self._cursor.pos

        if self._last_x is None or y == x == 0:
            self._last_x = x

        if y > 0:
            y = max(0, y - n)
            x = self._fix_x(y, min(self._last_x, self._line_widths[y]))
        else:
            x = 0

        self.cursor = y, x

    def move_cursor_down(self, n: int = 1):
        """Move cursor down `n` rows."""
        y, x = self._cursor.pos
        ey, ex = self.end_text_point

        if self._last_x is None or y == ey and x == ex:
            self._last_x = x

        if y < ey:
            y = min(ey, y + n)
            x = self._fix_x(y, min(self._last_x, self._line_widths[y]))
        else:
            x = ex

        self.cursor = y, x

    def move_word_left(self):
        """Move cursor a word left."""
        self._last_x = None
        last_x = self.cursor.x
        first_char_found = False
        while True:
            self.move_cursor_left()
            if self.cursor.x == last_x:
                break

            last_x = self.cursor.x

            current_char = egc_chr(self._pad.canvas[self.cursor]["ord"])
            if not first_char_found:
                if not current_char.isspace():
                    first_char_found = True
                    char_is_word_char = is_word_char(current_char)
            elif current_char.isspace() or char_is_word_char != is_word_char(
                current_char
            ):
                self.move_cursor_right()
                break

    def move_word_right(self):
        """Move cursor a word right."""
        self._last_x = None
        last_x = self.cursor.x
        first_char_found = False
        while True:
            self.move_cursor_right()
            if self.cursor.x == last_x:
                break

            last_x = self.cursor.x

            current_char = egc_chr(self._pad.canvas[self.cursor]["ord"])
            if not first_char_found:
                if not current_char.isspace():
                    first_char_found = True
                    char_is_word_char = is_word_char(current_char)
            elif current_char.isspace() or char_is_word_char != is_word_char(
                current_char
            ):
                break

    def _enter(self):
        self._move_undo_buffer_to_stack()
        undos = []
        if undo := self.delete_selection():
            undos.append(undo)
        undos.append(self._add_text(self.cursor, "\n"))
        self._undo_stack.append(undos)
        self._redo_stack.clear()

    def _tab(self):
        self._move_undo_buffer_to_stack()
        undos = []
        if undo := self.delete_selection():
            undos.append(undo)
        undos.append(self._add_text(self.cursor, "    "))
        self._undo_stack.append(undos)
        self._redo_stack.clear()

    def _backspace(self):
        if self.has_nonempty_selection:
            self._move_undo_buffer_to_stack("del")
            self._undo_buffer.append(self.delete_selection())
        else:
            if self._undo_buffer_type != "del":
                self._move_undo_buffer_to_stack("del")

            end = self.cursor
            self.move_cursor_left()
            start = self.cursor
            self.cursor = end
            if start != end:
                self._undo_buffer.append(self._del_text(start, end))

    def _delete(self):
        if self.has_nonempty_selection:
            self._move_undo_buffer_to_stack("del")
            self._undo_buffer.append(self.delete_selection())
        else:
            if self._undo_buffer_type != "del":
                self._move_undo_buffer_to_stack("del")

            start = self.cursor
            self.move_cursor_right()
            end = self.cursor
            self.cursor = start
            if start != end:
                self._undo_buffer.append(self._del_text(start, end))

    def _left(self):
        if self.has_nonempty_selection:
            select_start = min(self._selection_start, self._selection_end)
            self.unselect()
            self.cursor = select_start
        else:
            self.unselect()
            self.move_cursor_left()

    def _right(self):
        if self.has_nonempty_selection:
            select_end = max(self._selection_start, self._selection_end)
            self.unselect()
            self.cursor = select_end
        else:
            self.unselect()
            self.move_cursor_right()

    def _ctrl_left(self):
        self.unselect()
        self.move_word_left()

    def _ctrl_right(self):
        self.unselect()
        self.move_word_right()

    def _ctrl_a(self):
        """Select all."""
        self._selection_start = 0, 0
        self._selection_end = self.end_text_point
        self.cursor = self.end_text_point

    def _ctrl_d(self):
        """Select word."""
        self.unselect()
        last_x = self.cursor.x
        while True:
            self.move_cursor_left()
            if last_x == self.cursor.x:
                break
            if not is_word_char(egc_chr(self._pad.canvas[self.cursor]["ord"])):
                self.move_cursor_right()
                break
            last_x = self.cursor.x

        self.select()
        last_x = self.cursor.x
        while True:
            if not is_word_char(egc_chr(self._pad.canvas[self.cursor]["ord"])):
                break
            self.move_cursor_right()
            if last_x == self.cursor.x:
                break
            last_x = self.cursor.x

    def _up(self):
        if self.is_selecting:
            select_start = min(self._selection_start, self._selection_end)
            self.unselect()
            self.cursor = select_start
        self.move_cursor_up()

    def _down(self):
        if self.is_selecting:
            select_end = max(self._selection_start, self._selection_end)
            self.unselect()
            self.cursor = select_end
        self.move_cursor_down()

    def _pgup(self):
        if self.is_selecting:
            select_start = min(self._selection_start, self._selection_end)
            self.unselect()
            self.cursor = select_start
        self.move_cursor_up(self.page_lines)

    def _pgdn(self):
        if self.is_selecting:
            select_end = max(self._selection_start, self._selection_end)
            self.unselect()
            self.cursor = select_end
        self.move_cursor_down(self.page_lines)

    def _home(self):
        self.unselect()
        self._last_x = None
        self.cursor = self.cursor.y, 0

    def _end(self):
        self.unselect()
        self._last_x = None
        y = self.cursor.y
        self.cursor = y, self._line_widths[y]

    def _shift_left(self):
        self.select()
        self.move_cursor_left()

    def _shift_right(self):
        self.select()
        self.move_cursor_right()

    def _shift_ctrl_left(self):
        self.select()
        self.move_word_left()

    def _shift_ctrl_right(self):
        self.select()
        self.move_word_right()

    def _shift_up(self):
        self.select()
        self.move_cursor_up()

    def _shift_down(self):
        self.select()
        self.move_cursor_down()

    def _shift_pgup(self):
        self.select()
        self.move_cursor_up(self.page_lines)

    def _shift_pgdn(self):
        self.select()
        self.move_cursor_down(self.page_lines)

    def _shift_home(self):
        self.select()
        self._last_x = None
        self.cursor = self.cursor.y, 0

    def _shift_end(self):
        self.select()
        self._last_x = None
        y = self.cursor.y
        self.cursor = y, self._line_widths[y]

    def _escape(self):
        if self.has_nonempty_selection:
            self.unselect()
            self._highlight_selection()
        else:
            self.unselect()
            self.blur()

    def _ascii(self, key):
        if self.has_nonempty_selection:
            self._move_undo_buffer_to_stack("add")
            self._undo_buffer.append(self.delete_selection())
        elif self._undo_buffer_type != "add":
            self._move_undo_buffer_to_stack("add")
        self._undo_buffer.append(self._add_text(self.cursor, key))

    __HANDLERS = {
        ("enter", False, False, False): _enter,
        ("tab", False, False, False): _tab,
        ("backspace", False, False, False): _backspace,
        ("delete", False, False, False): _delete,
        ("left", False, False, False): _left,
        ("right", False, False, False): _right,
        ("left", False, True, False): _ctrl_left,
        ("right", False, True, False): _ctrl_right,
        ("up", False, False, False): _up,
        ("down", False, False, False): _down,
        ("page_up", False, False, False): _pgup,
        ("page_down", False, False, False): _pgdn,
        ("home", False, False, False): _home,
        ("end", False, False, False): _end,
        ("left", False, False, True): _shift_left,
        ("right", False, False, True): _shift_right,
        ("left", False, True, True): _shift_ctrl_left,
        ("right", False, True, True): _shift_ctrl_right,
        ("up", False, False, True): _shift_up,
        ("down", False, False, True): _shift_down,
        ("page_up", False, False, True): _shift_pgup,
        ("page_down", False, False, True): _shift_pgdn,
        ("home", False, False, True): _shift_home,
        ("end", False, False, True): _shift_end,
        ("escape", False, False, False): _escape,
        ("z", False, True, False): undo,
        ("y", False, True, False): redo,
        ("r", False, True, False): redo,
        ("a", False, True, False): _ctrl_a,
        ("d", False, True, False): _ctrl_d,
    }

    def on_key(self, key_event: KeyEvent) -> bool | None:
        """Process key press."""
        if not self.is_focused:
            return super().on_key(key_event)

        if (
            not key_event.alt
            and not key_event.ctrl
            and not key_event.shift
            and len(key_event.key) == 1
        ):
            self._ascii(key_event.key)
        elif handler := self.__HANDLERS.get(astuple(key_event)):
            handler(self)
        else:
            return super().on_key(key_event)

        return True

    def on_paste(self, paste_event: PasteEvent) -> bool | None:
        """Add paste to text pad."""
        if not self.is_focused:
            return

        self._move_undo_buffer_to_stack()
        undos = []
        if undo := self.delete_selection():
            undos.append(undo)
        undos.append(self._add_text(self.cursor, paste_event.paste))
        self._undo_stack.append(undos)
        self._redo_stack.clear()

        return True

    def grab(self, mouse_event):
        """Start selection on grab."""
        if mouse_event.button == "left" and self._pad.collides_point(mouse_event.pos):
            super().grab(mouse_event)

            y, x = self._pad.to_local(mouse_event.pos)
            if y >= len(self._line_widths):
                return

            x = min(x, self._line_widths[y])
            if not mouse_event.shift:
                self.unselect()

            self.cursor = y, x
            self.select()  # Need at least an empty selection for `grab_update`.

    def grab_update(self, mouse_event: MouseEvent):
        """Update selection on grab update."""
        if self._pad.collides_point(mouse_event.pos):
            y, x = self._pad.to_local(mouse_event.pos)
            if y < len(self._line_widths):
                x = min(x, self._line_widths[y])
                self.cursor = y, x
        else:
            cy, cx = self.cursor
            y, x = self.to_local(mouse_event.pos)
            h, w = self.size

            if y < 0:
                self.move_cursor_up()
            elif y >= h:
                self.move_cursor_down()

            if x < 0:
                if cx > 0:
                    self.move_cursor_left()
            elif x >= w:
                if cx < self._line_widths[cy]:
                    self.move_cursor_right()

    def ungrab(self, mouse_event):
        """Clear an empty selection on ungrab."""
        super().ungrab(mouse_event)
        if not self.has_nonempty_selection:
            self.unselect()



"""Color data structures."""

import re
from typing import Final, NamedTuple, Self, TypedDict

__all__ = ["AColor", "Color", "ColorTheme"]

_HEXCODE_RE: Final = re.compile(r"^#?[0-9a-fA-F]{6}$")
_AHEXCODE_RE: Final = re.compile(r"^#?[0-9a-fA-F]{8}$")

type Hexcode = str
"""
A hexcode string for 24-bit color. Six hex digits optionally preceded by "#", e.g.,
"#abc123".
"""

type AHexcode = str
"""
A hexcode string for a 32-bit color. Six or eight hex digits optionally preceded by
"#", e.g., "#abcd1234". If eight hex digits are given, the last two digits specify the
alpha channel, otherwise, the alpha channel value defaults to 255 so that the six digit
string ``"#abc123`` is equivalent to ``"#abc123ff"``.
"""


def validate_hexcode(hexcode: Hexcode) -> bool:
    return bool(_HEXCODE_RE.match(hexcode))


def validate_ahexcode(ahexcode: AHexcode) -> bool:
    return validate_hexcode(ahexcode) or bool(_AHEXCODE_RE.match(ahexcode))


def _to_hex(channel: int):
    return f"{channel:02x}"


class Color(NamedTuple):
    """
    A 24-bit color.

    Parameters
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.

    Attributes
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.

    Methods
    -------
    from_hex(hexcode)
        Create a :class:`Color` from a hex code.
    to_hex()
        Return color's hexcode.
    """

    red: int
    green: int
    blue: int

    @classmethod
    def from_hex(cls, hexcode: Hexcode) -> Self:
        """
        Create a :class:`Color` from a hex code.

        Parameters
        ----------
        hexcode : Hexcode
            A color hex code.

        Returns
        -------
        Color
            A new color.
        """
        if not validate_hexcode(hexcode):
            raise ValueError(f"{hexcode!r} is not a valid hex code.")

        digits = hexcode.removeprefix("#")
        return cls(int(digits[:2], 16), int(digits[2:4], 16), int(digits[4:], 16))

    def to_hex(self) -> Hexcode:
        """
        Return color's hexcode.

        Returns
        -------
        Hexcode
            The hexcode of the color.
        """
        hexcode = "".join(_to_hex(channel) for channel in self)
        return f"#{hexcode}"


class AColor(NamedTuple):
    """
    A 32-bit color with an alpha channel.

    Parameters
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.
    alpha : int, default: 255
        The alpha component.

    Attributes
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.
    alpha : int
        The alpha component.

    Methods
    -------
    from_hex(hexcode)
        Create an :class:`AColor` from a hex code.
    to_hex()
        Return color's hexcode.
    """

    red: int
    green: int
    blue: int
    alpha: int = 255

    @classmethod
    def from_hex(cls, hexcode: AHexcode) -> Self:
        """
        Create an :class:`AColor` from a hex code.

        If the hexcode only has six hex digits, the alpha channel's value will default
        to 255.

        Parameters
        ----------
        hexcode : AHexcode
            A color hex code.

        Returns
        -------
        AColor
            A new color with an alpha channel.
        """
        if not validate_ahexcode(hexcode):
            raise ValueError(f"{hexcode!r} is not a valid hex code")

        digits = hexcode.removeprefix("#")
        return cls(
            int(digits[:2], 16),
            int(digits[2:4], 16),
            int(digits[4:6], 16),
            int(digits[6:] or "ff", 16),
        )

    def to_hex(self) -> AHexcode:
        """
        Return color's hexcode.

        Returns
        -------
        AHexcode
            The hexcode of the color.
        """
        hexcode = "".join(_to_hex(channel) for channel in self)
        return f"#{hexcode}"


class ColorTheme(TypedDict, total=False):
    """
    Colors for themable gadgets.

    Missing colors will use the default color theme.

    Attributes
    ----------
    primary_fg : Hexcode
        The primary foreground color.
    primary_bg : Hexcode
        The primary background color.
    text_pad_line_highlight_fg: Hexcode
        Text pad line highlight foreground color.
    text_pad_line_highlight_bg: Hexcode
        Text pad line highlight background color.
    text_pad_selection_highlight_fg : Hexcode
        Text pad selection foreground color.
    text_pad_selection_highlight_bg : Hexcode
        Text pad selection background color.
    textbox_primary_fg : Hexcode
        Text pad primary foreground color.
    textbox_primary_bg : Hexcode
        Text pad primary background color.
    textbox_selection_highlight_fg : Hexcode
        Textbox selection foreground color.
    textbox_selection_highlight_bg : Hexcode
        Textbox selection background color.
    textbox_placeholder_fg : Hexcode
        Textbox placeholder foreground color.
    textbox_placeholder_bg : Hexcode
        Textbox placeholder background color.
    button_normal_fg : Hexcode
        Button normal foreground color.
    button_normal_bg : Hexcode
        Button normal background color.
    button_hover_fg : Hexcode
        Button hover foreground color.
    button_hover_bg : Hexcode
        Button hover background color.
    button_press_fg : Hexcode
        Button press foreground color.
    button_press_bg : Hexcode
        Button press background color.
    button_disallowed_fg : Hexcode
        Button disallowed foreground color.
    button_disallowed_bg : Hexcode
        Button disallowed background color.
    menu_item_hover_fg : Hexcode
        Menu item hover foreground color.
    menu_item_hover_bg : Hexcode
        Menu item hover background color.
    menu_item_selected_fg : Hexcode
        Menu item selected foreground color.
    menu_item_selected_bg : Hexcode
        Menu item selected background color.
    menu_item_disallowed_fg : Hexcode
        Menu item disallowed foreground color.
    menu_item_disallowed_bg : Hexcode
        Menu item disallowed background color.
    titlebar_normal_fg : Hexcode
        Titlebar normal foreground color.
    titlebar_normal_bg : Hexcode
        Titlebar normal background color.
    titlebar_inactive_fg : Hexcode
        Titlebar inactive foreground color.
    titlebar_inactive_bg : Hexcode
        Titlebar inactive background color.
    data_table_sort_indicator_fg : Hexcode
        Data table sort indicator foreground color.
    data_table_sort_indicator_bg : Hexcode
        Data table sort indicator background color.
    data_table_hover_fg : Hexcode
        Data table hover foreground color.
    data_table_hover_bg : Hexcode
        Data table hover background color.
    data_table_stripe_fg : Hexcode
        Data table stripe foreground color.
    data_table_stripe_bg : Hexcode
        Data table stripe background color.
    data_table_stripe_hover_fg : Hexcode
        Data table stripe hover foreground color.
    data_table_stripe_hover_bg : Hexcode
        Data table stripe hover background color.
    data_table_selected_fg : Hexcode
        Data table selected foreground color.
    data_table_selected_bg : Hexcode
        Data table selected background color.
    data_table_selected_hover_fg : Hexcode
        Data table selected hover foreground color.
    data_table_selected_hover_bg : Hexcode
        Data table selected hover background color.
    progress_bar_fg : Hexcode
        Progress bar foreground color.
    progress_bar_bg : Hexcode
        Progress bar background color.
    markdown_link_fg : Hexcode
        Markdown link foreground color.
    markdown_link_bg : Hexcode
        Markdown link background color.
    markdown_link_hover_fg : Hexcode
        Markdown link hover foreground color.
    markdown_link_hover_bg : Hexcode
        Markdown link hover background color.
    markdown_inline_code_fg : Hexcode
        Markdown inline code foreground color.
    markdown_inline_code_bg : Hexcode
        Markdown inline code background color.
    markdown_quote_fg : Hexcode
        Markdown quote foreground color.
    markdown_quote_bg : Hexcode
        Markdown quote background color.
    markdown_title_fg : Hexcode
        Markdown title foreground color.
    markdown_title_bg : Hexcode
        Markdown title background color.
    markdown_image_fg : Hexcode
        Markdown image foreground color.
    markdown_image_bg : Hexcode
        Markdown image background color.
    markdown_block_code_background : Hexcode
        Markdown block code background color.
    markdown_quote_block_code_background : Hexcode
        Markdown quote block code background color.
    markdown_header_background : Hexcode
        Markdown header background color.
    scroll_view_scrollbar : Hexcode
        Scroll view scrollbar color.
    scroll_view_indicator_normal : Hexcode
        Scroll view indicator normal color.
    scroll_view_indicator_hover : Hexcode
        Scroll view indicator hover color.
    scroll_view_indicator_press : Hexcode
        Scroll view indicator press color.
    window_border_normal : Hexcode
        Window border normal color.
    window_border_inactive : Hexcode
        Window border inactive color.
    """

    primary_fg: Hexcode
    """The primary foreground color."""
    primary_bg: Hexcode
    """The primary background color."""
    text_pad_line_highlight_fg: Hexcode
    """Text pad line highlight foreground color."""
    text_pad_line_highlight_bg: Hexcode
    """Text pad line highlight background color."""
    text_pad_selection_highlight_fg: Hexcode
    """Text pad selection foreground color."""
    text_pad_selection_highlight_bg: Hexcode
    """Text pad selection background color."""
    textbox_primary_fg: Hexcode
    """Text pad primary foreground color."""
    textbox_primary_bg: Hexcode
    """Text pad primary background color."""
    textbox_selection_highlight_fg: Hexcode
    """Textbox selection foreground color."""
    textbox_selection_highlight_bg: Hexcode
    """Textbox selection background color."""
    textbox_placeholder_fg: Hexcode
    """Textbox placeholder foreground color."""
    textbox_placeholder_bg: Hexcode
    """Textbox placeholder background color."""
    button_normal_fg: Hexcode
    """Button normal foreground color."""
    button_normal_bg: Hexcode
    """Button normal background color."""
    button_hover_fg: Hexcode
    """Button hover foreground color."""
    button_hover_bg: Hexcode
    """Button hover background color."""
    button_press_fg: Hexcode
    """Button press foreground color."""
    button_press_bg: Hexcode
    """Button press background color."""
    button_disallowed_fg: Hexcode
    """Button disallowed foreground color."""
    button_disallowed_bg: Hexcode
    """Button disallowed background color."""
    menu_item_hover_fg: Hexcode
    """Menu item hover foreground color."""
    menu_item_hover_bg: Hexcode
    """Menu item hover background color."""
    menu_item_selected_fg: Hexcode
    """Menu item selected foreground color."""
    menu_item_selected_bg: Hexcode
    """Menu item selected background color."""
    menu_item_disallowed_fg: Hexcode
    """Menu item disallowed foreground color."""
    menu_item_disallowed_bg: Hexcode
    """Menu item disallowed background color."""
    titlebar_normal_fg: Hexcode
    """Titlebar normal foreground color."""
    titlebar_normal_bg: Hexcode
    """Titlebar normal background color."""
    titlebar_inactive_fg: Hexcode
    """Titlebar inactive foreground color."""
    titlebar_inactive_bg: Hexcode
    """Titlebar inactive background color."""
    data_table_sort_indicator_fg: Hexcode
    """Data table sort indicator foreground color."""
    data_table_sort_indicator_bg: Hexcode
    """Data table sort indicator background color."""
    data_table_hover_fg: Hexcode
    """Data table hover foreground color."""
    data_table_hover_bg: Hexcode
    """Data table hover background color."""
    data_table_stripe_fg: Hexcode
    """Data table stripe foreground color."""
    data_table_stripe_bg: Hexcode
    """Data table stripe background color."""
    data_table_stripe_hover_fg: Hexcode
    """Data table stripe hover foreground color."""
    data_table_stripe_hover_bg: Hexcode
    """Data table stripe hover background color."""
    data_table_selected_fg: Hexcode
    """Data table selected foreground color."""
    data_table_selected_bg: Hexcode
    """Data table selected background color."""
    data_table_selected_hover_fg: Hexcode
    """Data table selected hover foreground color."""
    data_table_selected_hover_bg: Hexcode
    """Data table selected hover background color."""
    progress_bar_fg: Hexcode
    """Progress bar foreground color."""
    progress_bar_bg: Hexcode
    """Progress bar background color."""
    markdown_link_fg: Hexcode
    """Markdown link foreground color."""
    markdown_link_bg: Hexcode
    """Markdown link background color."""
    markdown_link_hover_fg: Hexcode
    """Markdown link hover foreground color."""
    markdown_link_hover_bg: Hexcode
    """Markdown link hover background color."""
    markdown_inline_code_fg: Hexcode
    """Markdown inline code foreground color."""
    markdown_inline_code_bg: Hexcode
    """Markdown inline code background color."""
    markdown_quote_fg: Hexcode
    """Markdown quote foreground color."""
    markdown_quote_bg: Hexcode
    """Markdown quote background color."""
    markdown_title_fg: Hexcode
    """Markdown title foreground color."""
    markdown_title_bg: Hexcode
    """Markdown title background color."""
    markdown_image_fg: Hexcode
    """Markdown image foreground color."""
    markdown_image_bg: Hexcode
    """Markdown image background color."""
    markdown_block_code_bg: Hexcode
    """Markdown block code background color."""
    markdown_quote_block_code_bg: Hexcode
    """Markdown quote block code background color."""
    markdown_header_bg: Hexcode
    """Markdown header background color."""
    scroll_view_scrollbar: Hexcode
    """Scroll view scrollbar color."""
    scroll_view_indicator_normal: Hexcode
    """Scroll view indicator normal color."""
    scroll_view_indicator_hover: Hexcode
    """Scroll view indicator hover color."""
    scroll_view_indicator_press: Hexcode
    """Scroll view indicator press color."""
    window_border_normal: Hexcode
    """Window border normal color."""
    window_border_inactive: Hexcode
    """Window border inactive color."""
